"""
Test Serverless Job Module
"""

from unittest.mock import Mock, patch

from unittest import IsolatedAsyncioTestCase
from aiohttp import ClientResponse, ClientResponseError
from aiohttp.test_utils import make_mocked_coro

from runpod.http_client import TooManyRequests
from runpod.serverless.modules import rp_job


class TestJob(IsolatedAsyncioTestCase):
    """Tests for the get_job function."""

    async def test_get_job_200(self):
        """Tests the get_job function with a valid 200 response."""
        # Mock the 200 response
        response = Mock(ClientResponse)
        response.status = 200
        response.content_type = "application/json"
        response.content_length = 50
        response.json = make_mocked_coro(
            return_value={"id": "123", "input": {"number": 1}}
        )

        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            job = await rp_job.get_job(mock_session)
            # Assertions for the success case
            self.assertEqual(job, [{"id": "123", "input": {"number": 1}}])

    async def test_get_job_204(self):
        """Tests the get_job function with a 204 response."""
        # Mock 204 No Content response
        response = Mock(ClientResponse)
        response.status = 204
        response.content_type = "application/json"
        response.content_length = 0

        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            job = await rp_job.get_job(mock_session)
            self.assertIsNone(job)
            self.assertEqual(mock_session.get.call_count, 1)

    async def test_get_job_400(self):
        """Tests the get_job function with a 400 response."""
        # Mock 400 response
        response = Mock(ClientResponse)
        response.status = 400

        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            job = await rp_job.get_job(mock_session)
            self.assertIsNone(job)

    async def test_get_job_429(self):
        """Tests the get_job function with a 429 response."""
        response = Mock(ClientResponse)
        response.raise_for_status.side_effect = TooManyRequests(
            request_info=None,
            history=(),
            status=429,
        )

        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            with self.assertRaises(ClientResponseError):
                await rp_job.get_job(mock_session)

    async def test_get_job_500(self):
        """Tests the get_job function with a 500 response."""
        # Mock 500 response
        response = Mock(ClientResponse)
        response.raise_for_status.side_effect = TooManyRequests(
            request_info=None,  # Not needed for the test
            history=(),  # Not needed for the test
            status=500, 
        )
        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            with self.assertRaises(Exception):
                await rp_job.get_job(mock_session)

    async def test_get_job_no_id(self):
        """Tests the get_job function with a 200 response but no 'id' field."""
        response = Mock(ClientResponse)
        response.status = 200
        response.content_type = "application/json"
        response.content_length = 50
        response.json = make_mocked_coro(return_value={"input": "foobar"})

        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            with self.assertRaises(Exception) as context:
                await rp_job.get_job(mock_session)
            self.assertEqual(str(context.exception), "Job has missing field(s): id or input.")

    async def test_get_job_invalid_content_type(self):
        """Tests the get_job function with an invalid content type."""
        response = Mock(ClientResponse)
        response.status = 200
        response.content_type = "text/html"  # Invalid content type
        response.content_length = 50

        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            job = await rp_job.get_job(mock_session)
            self.assertIsNone(job)

    async def test_get_job_empty_content(self):
        """Tests the get_job function with an empty content response."""
        response = Mock(ClientResponse)
        response.status = 200
        response.content_type = "application/json"
        response.content_length = 0  # No content to parse

        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.return_value = response
            job = await rp_job.get_job(mock_session)
            self.assertIsNone(job)

    async def test_get_job_exception(self):
        """Tests the get_job function with a raised exception."""
        with patch("aiohttp.ClientSession") as mock_session, patch(
            "runpod.serverless.modules.rp_job.JOB_GET_URL", "http://mock.url"
        ):
            mock_session.get.return_value.__aenter__.side_effect = Exception("Unexpected error")
            with self.assertRaises(Exception) as context:
                await rp_job.get_job(mock_session)
            self.assertEqual(str(context.exception), "Unexpected error")


class TestRunJob(IsolatedAsyncioTestCase):
    """Tests the run_job function"""

    async def asyncSetUp(self) -> None:
        self.sample_job = {
            "id": "123",
            "input": {
                "test_input": None,
            },
        }

    async def test_simple_job(self):
        """
        Tests the run_job function
        """
        mock_handler = Mock()

        mock_handler.return_value = "test"
        job_result = await rp_job.run_job(mock_handler, self.sample_job)
        assert job_result == {"output": "test"}

        mock_handler.return_value = ["test1", "test2"]
        job_result_list = await rp_job.run_job(mock_handler, self.sample_job)
        assert job_result_list == {"output": ["test1", "test2"]}

        mock_handler.return_value = 123
        job_result_int = await rp_job.run_job(mock_handler, self.sample_job)
        assert job_result_int == {"output": 123}

    async def test_job_with_errors(self):
        """
        Tests the run_job function with errors
        """
        mock_handler = Mock()
        mock_handler.return_value = {"error": "test"}

        job_result = await rp_job.run_job(mock_handler, self.sample_job)

        assert job_result == {"error": "test"}

    async def test_job_with_raised_exception(self):
        """
        Tests the run_job function with a raised exception
        """
        mock_handler = Mock()
        mock_handler.side_effect = Exception

        job_result = await rp_job.run_job(mock_handler, self.sample_job)

        assert "error" in job_result

    async def test_job_with_refresh_worker(self):
        """
        Tests the run_job function with refresh_worker
        """
        mock_handler = Mock()
        mock_handler.return_value = {"refresh_worker": True}

        job_result = await rp_job.run_job(mock_handler, self.sample_job)

        assert job_result["stopPod"] is True

    async def test_job_bool_output(self):
        """
        Tests the run_job function with a boolean output
        """
        mock_handler = Mock()
        mock_handler.return_value = True

        job_result = await rp_job.run_job(mock_handler, self.sample_job)

        assert job_result == {"output": True}

    async def test_job_with_exception(self):
        """
        Tests the run_job function with an exception
        """
        mock_handler = Mock()
        mock_handler.side_effect = Exception

        job_result = await rp_job.run_job(mock_handler, self.sample_job)

        self.assertRaises(Exception, job_result)


class TestRunJobGenerator(IsolatedAsyncioTestCase):
    """Tests the run_job_generator function"""

    def handler_gen_success(self, job):  # pylint: disable=unused-argument
        """
        Test handler that returns a generator.
        """
        yield "partial_output_1"
        yield "partial_output_2"

    async def handler_async_gen_success(self, job):  # pylint: disable=unused-argument
        """
        Test handler that returns an async generator.
        """
        yield "partial_output_1"
        yield "partial_output_2"

    def handler_fail(self, job):
        """
        Test handler that raises an exception.
        """
        raise Exception("Test Exception")  # pylint: disable=broad-exception-raised

    async def test_run_job_generator_success(self):
        """
        Tests the run_job_generator function with a successful generator
        """
        handler = self.handler_gen_success
        job = {"id": "123"}

        with patch(
            "runpod.serverless.modules.rp_job.log", new_callable=Mock
        ) as mock_log:
            result = [i async for i in rp_job.run_job_generator(handler, job)]

        assert result == [
            {"output": "partial_output_1"},
            {"output": "partial_output_2"},
        ]
        assert mock_log.error.call_count == 0
        assert mock_log.info.call_count == 1
        mock_log.info.assert_called_with("Finished running generator.", "123")

    async def test_run_job_generator_success_async(self):
        """
        Tests the run_job_generator function with a successful generator
        """
        handler = self.handler_async_gen_success
        job = {"id": "123"}

        with patch(
            "runpod.serverless.modules.rp_job.log", new_callable=Mock
        ) as mock_log:
            result = [i async for i in rp_job.run_job_generator(handler, job)]

        assert result == [
            {"output": "partial_output_1"},
            {"output": "partial_output_2"},
        ]
        assert mock_log.error.call_count == 0
        assert mock_log.info.call_count == 1
        mock_log.info.assert_called_with("Finished running generator.", "123")

    async def test_run_job_generator_exception(self):
        """
        Tests the run_job_generator function with an exception
        """
        handler = self.handler_fail
        job = {"id": "123"}

        with patch(
            "runpod.serverless.modules.rp_job.log", new_callable=Mock
        ) as mock_log:
            result = [i async for i in rp_job.run_job_generator(handler, job)]

        assert len(result) == 1
        assert "error" in result[0]
        assert mock_log.error.call_count == 1
        assert mock_log.info.call_count == 1
        mock_log.info.assert_called_with("Finished running generator.", "123")



""" Tests for environment variables module """

import os
import unittest

from runpod.serverless.modules.worker_state import (
    Job,
    JobsProgress,
    IS_LOCAL_TEST,
    WORKER_ID,
)


class TestEnvVars(unittest.TestCase):
    """Tests for environment variables module"""

    def setUp(self):
        """
        Set up test variables
        """
        self.test_api_key = "test_api_key"
        os.environ["RUNPOD_AI_API_KEY"] = self.test_api_key

    def test_is_local_test(self):
        """
        Tests if IS_LOCAL_TEST flag is properly set
        """
        os.environ.pop("RUNPOD_WEBHOOK_GET_JOB", None)
        self.assertEqual(IS_LOCAL_TEST, True)

    def test_worker_id(self):
        """
        Tests if WORKER_ID is properly set
        """
        os.environ["RUNPOD_POD_ID"] = WORKER_ID

        self.assertEqual(WORKER_ID, os.environ.get("RUNPOD_POD_ID"))


class TestJob(unittest.TestCase):

    def test_initialization_with_basic_attributes(self):
        """Test basic initialization of Job object."""
        job = Job(id="job_123", input={"task": "data_process"}, webhook="http://example.com/webhook")
        self.assertEqual(job.id, "job_123")
        self.assertEqual(job.input, {"task": "data_process"})
        self.assertEqual(job.webhook, "http://example.com/webhook")

    def test_initialization_with_additional_kwargs(self):
        """Test initialization with extra kwargs dynamically creating attributes."""
        job = Job(id="job_456", status="pending", priority=5)
        self.assertEqual(job.id, "job_456")
        self.assertEqual(job.status, "pending")
        self.assertEqual(job.priority, 5)

    def test_equality(self):
        """Test equality between two Job objects based on the job ID."""
        job1 = Job(id="job_123")
        job2 = Job(id="job_123")
        job3 = Job(id="job_456")
        
        self.assertEqual(job1, job2)
        self.assertNotEqual(job1, job3)

    def test_hashing(self):
        """Test hashing of Job object based on the job ID."""
        job1 = Job(id="job_123")
        job2 = Job(id="job_123")
        job3 = Job(id="job_456")
        
        self.assertEqual(hash(job1), hash(job2))
        self.assertNotEqual(hash(job1), hash(job3))

    def test_string_representation(self):
        """Test the string representation of the Job object."""
        job = Job(id="job_123")
        self.assertEqual(str(job), "job_123")

    def test_none_input(self):
        """Test initialization with None values."""
        job = Job(id="job_123", input=None, webhook=None)
        self.assertEqual(job.id, "job_123")
        self.assertIsNone(job.input)
        self.assertIsNone(job.webhook)

    def test_dynamic_kwargs_assignment(self):
        """Test if kwargs are dynamically assigned as attributes."""
        job = Job(id="job_789", foo="bar", custom_attr=42)
        self.assertEqual(job.foo, "bar")
        self.assertEqual(job.custom_attr, 42)

    def test_missing_attributes(self):
        """Test that accessing non-existent attributes raises AttributeError."""
        job = Job(id="job_123")
        with self.assertRaises(AttributeError):
            _ = job.non_existent_attr


class TestJobsProgress(unittest.IsolatedAsyncioTestCase):
    """Tests for JobsProgress class"""

    async def asyncSetUp(self):
        """
        Set up test variables
        """
        self.jobs = JobsProgress()
        self.jobs.clear()  # clear jobs before each test

    def test_singleton(self):
        jobs2 = JobsProgress()
        self.assertEqual(self.jobs, jobs2)

    async def test_add_job(self):
        assert not self.jobs.get_job_count()

        id = "123"
        self.jobs.add({"id": id})
        assert self.jobs.get_job_count() == 1

        job1 = self.jobs.get(id)
        assert job1 in self.jobs

        id = "234"
        self.jobs.add(id)
        assert self.jobs.get_job_count() == 2

        job2 = self.jobs.get(id)
        assert job2 in self.jobs

    async def test_remove_job(self):
        assert not self.jobs.get_job_count()

        job = {"id": "123"}
        self.jobs.add(job)
        assert self.jobs.get_job_count()

        self.jobs.remove("123")
        assert not self.jobs.get_job_count()

    async def test_get_job(self):
        for id in ["123", "234", "345"]:
            self.jobs.add({"id": id})

        job1 = self.jobs.get(id)
        assert job1 in self.jobs

    async def test_get_job_list(self):
        assert self.jobs.get_job_list() is None

        job1 = {"id": "123"}
        self.jobs.add(job1)

        job2 = {"id": "456"}
        self.jobs.add(job2)

        assert self.jobs.get_job_count() == 2
        assert self.jobs.get_job_list() in ["123,456", "456,123"]

    async def test_get_job_count(self):
        # test job count contention when adding and removing jobs in parallel
        pass


"""
Contains the main entrypoint for the RunPod Serverless Worker.

Arguments can be passed in when the worker is started, and will be passed to the worker.
"""

import argparse
import json
import os
import signal
import sys
import time
from typing import Any, Dict

from runpod.serverless import core

from ..version import __version__ as runpod_version
from . import worker
from .modules import rp_fastapi
from .modules.rp_logger import RunPodLogger
from .modules.rp_progress import progress_update

log = RunPodLogger()


# ---------------------------------------------------------------------------- #
#                              Run Time Arguments                              #
# ---------------------------------------------------------------------------- #
# Arguments will be passed in with the config under the key "rp_args"
parser = argparse.ArgumentParser(
    prog="runpod", description="Runpod Serverless Worker Arguments."
)
parser.add_argument(
    "--rp_log_level",
    type=str,
    default=None,
    help="""Controls what level of logs are printed to the console.
                    Options: ERROR, WARN, INFO, and DEBUG.""",
)

parser.add_argument(
    "--rp_debugger",
    action="store_true",
    default=None,
    help="Flag to enable the Debugger.",
)

# Hosted API
parser.add_argument(
    "--rp_serve_api",
    action="store_true",
    default=None,
    help="Flag to start the API server.",
)
parser.add_argument(
    "--rp_api_port", type=int, default=8000, help="Port to start the FastAPI server on."
)
parser.add_argument(
    "--rp_api_concurrency",
    type=int,
    default=1,
    help="Number of concurrent FastAPI workers.",
)
parser.add_argument(
    "--rp_api_host",
    type=str,
    default="localhost",
    help="Host to start the FastAPI server on.",
)

# Test input
parser.add_argument(
    "--test_input",
    type=str,
    default=None,
    help="Test input for the worker, formatted as JSON.",
)


def _set_config_args(config) -> dict:
    """
    Sets the config rp_args, removing any recognized arguments from sys.argv.
    Returns: config
    """
    args, unknown = parser.parse_known_args()
    sys.argv = [sys.argv[0]] + unknown

    # Directly assign the parsed arguments to config
    config["rp_args"] = vars(args)

    # Parse the test input from JSON
    if config["rp_args"]["test_input"]:
        config["rp_args"]["test_input"] = json.loads(config["rp_args"]["test_input"])

    # Parse the test output from JSON
    if config["rp_args"].get("test_output", None):
        config["rp_args"]["test_output"] = json.loads(config["rp_args"]["test_output"])

    # Set the log level
    if config["rp_args"]["rp_log_level"]:
        log.set_level(config["rp_args"]["rp_log_level"])

    return config


def _get_realtime_port() -> int:
    """
    Get the realtime port from the environment variable if it exists.
    """
    return int(os.environ.get("RUNPOD_REALTIME_PORT", "0"))


def _get_realtime_concurrency() -> int:
    """
    Get the realtime concurrency from the environment variable if it exists.
    """
    return int(os.environ.get("RUNPOD_REALTIME_CONCURRENCY", "1"))


def _signal_handler(sig, frame):
    """
    Handles the SIGINT signal.
    """
    del sig, frame
    log.info("SIGINT received. Shutting down.")
    sys.exit(0)


# ---------------------------------------------------------------------------- #
#                            Start Serverless Worker                           #
# ---------------------------------------------------------------------------- #
def start(config: Dict[str, Any]):
    """
    Starts the serverless worker.

    config (Dict[str, Any]): Configuration parameters for the worker.

    config["handler"] (Callable): The handler function to run.

    config["rp_args"] (Dict[str, Any]): Arguments for the worker, populated by runtime arguments.
    """
    print(f"--- Starting Serverless Worker |  Version {runpod_version} ---")

    signal.signal(signal.SIGINT, _signal_handler)

    config["reference_counter_start"] = time.perf_counter()
    config = _set_config_args(config)

    realtime_port = _get_realtime_port()
    realtime_concurrency = _get_realtime_concurrency()

    if config["rp_args"]["rp_serve_api"]:
        log.info("Starting API server.")
        api_server = rp_fastapi.WorkerAPI(config)

        api_server.start_uvicorn(
            api_host=config["rp_args"]["rp_api_host"],
            api_port=config["rp_args"]["rp_api_port"],
            api_concurrency=config["rp_args"]["rp_api_concurrency"],
        )
        return

    if realtime_port:
        log.info(f"Starting API server for realtime on port {realtime_port}.")
        api_server = rp_fastapi.WorkerAPI(config)

        api_server.start_uvicorn(
            api_host="0.0.0.0",
            api_port=realtime_port,
            api_concurrency=realtime_concurrency,
        )
        return

    # --------------------------------- SLS-Core --------------------------------- #

    if os.getenv("RUNPOD_SLS_CORE") is None and os.getenv("RUNPOD_USE_CORE") is not None:
            log.warn("RUNPOD_USE_CORE is deprecated. Please use RUNPOD_SLS_CORE instead.")
            core.main(config)
            return

    elif os.getenv("RUNPOD_SLS_CORE","false").lower() in ["1", "t", "true"]:
            core.main(config)
            return

    # --------------------------------- Standard --------------------------------- #
    worker.main(config)
    return



"""
runpod | serverless | rp_local.py
Provides the local testing functionality for runpod serverless worker.
"""

import json
import os
import sys
from typing import Any, Dict

from runpod.serverless.modules.rp_logger import RunPodLogger

from .rp_job import run_job

log = RunPodLogger()


async def run_local(config: Dict[str, Any]) -> None:
    """
    Runs the worker locally.
    """
    # Get the local test job
    if config["rp_args"].get("test_input", None):
        log.info("test_input set, using test_input as job input.")
        local_job = config["rp_args"]["test_input"]
    else:
        if not os.path.exists("test_input.json"):
            log.warn("test_input.json not found, exiting.")
            sys.exit(1)

        log.info("Using test_input.json as job input.")
        with open("test_input.json", "r", encoding="UTF-8") as file:
            local_job = json.loads(file.read())

    if local_job.get("input", None) is None:
        log.error("Job has no input parameter. Unable to run.")
        sys.exit(1)

    # Set the job ID
    local_job["id"] = local_job.get("id", "local_test")
    log.debug(f"Retrieved local job: {local_job}")

    job_result = await run_job(config["handler"], local_job)

    if job_result.get("error", None):
        log.error(f"Job {local_job['id']} failed with error: {job_result['error']}")
        sys.exit(1)

    log.info(f"Job {local_job['id']} completed successfully.")
    log.info(f"Job result: {job_result}")

    # Compare to sample output, if provided
    if config["rp_args"].get("test_output", None):
        log.info("test_output set, comparing output to test_output.")
        if job_result != config["rp_args"]["test_output"]:
            log.error("Job output does not match test_output.")
            sys.exit(1)
        log.info("Job output matches test_output.")

    log.info("Local testing complete, exiting.")
    sys.exit(0)



""" Example of getting the health of an endpoint. """

import runpod

# Set your global API key with `runpod config` or uncomment the line below:
# runpod.api_key = "YOUR_RUNPOD_API_KEY"

endpoint = runpod.Endpoint("gwp4kx5yd3nur1")

endpoint_health = endpoint.health()

print(endpoint_health)



"""
    This module is used to handle HTTP requests.
"""

import json
import os

from aiohttp import ClientError
from aiohttp_retry import FibonacciRetry, RetryClient

from runpod.http_client import ClientSession
from runpod.serverless.modules.rp_logger import RunPodLogger

from .worker_state import WORKER_ID

JOB_DONE_URL_TEMPLATE = str(
    os.environ.get("RUNPOD_WEBHOOK_POST_OUTPUT", "JOB_DONE_URL")
)
JOB_DONE_URL = JOB_DONE_URL_TEMPLATE.replace("$RUNPOD_POD_ID", WORKER_ID)

JOB_STREAM_URL_TEMPLATE = str(
    os.environ.get("RUNPOD_WEBHOOK_POST_STREAM", "JOB_STREAM_URL")
)
JOB_STREAM_URL = JOB_STREAM_URL_TEMPLATE.replace("$RUNPOD_POD_ID", WORKER_ID)

log = RunPodLogger()


async def _transmit(client_session: ClientSession, url, job_data):
    """
    Wrapper for transmitting results via POST.
    """
    retry_options = FibonacciRetry(attempts=3)
    retry_client = RetryClient(
        client_session=client_session, retry_options=retry_options
    )

    kwargs = {
        "data": job_data,
        "headers": {
            "charset": "utf-8",
            "Content-Type": "application/x-www-form-urlencoded",
        },
        "raise_for_status": True,
    }

    async with retry_client.post(url, **kwargs) as client_response:
        await client_response.text()


async def _handle_result(
    session: ClientSession, job_data, job, url_template, log_message, is_stream=False
):
    """
    A helper function to handle the result, either for sending or streaming.
    """
    try:
        session.headers["X-Request-ID"] = job["id"]

        serialized_job_data = json.dumps(job_data, ensure_ascii=False)

        is_stream = "true" if is_stream else "false"
        url = url_template.replace("$ID", job["id"]) + f"&isStream={is_stream}"

        await _transmit(session, url, serialized_job_data)
        log.debug(f"{log_message}", job["id"])

    except ClientError as err:
        log.error(f"Failed to return job results. | {err}", job["id"])

    except (TypeError, RuntimeError) as err:
        log.error(f"Error while returning job result. | {err}", job["id"])

    finally:
        # job_data status is used for local development with FastAPI
        if (
            url_template == JOB_DONE_URL
            and job_data.get("status", None) != "IN_PROGRESS"
        ):
            log.info("Finished.", job["id"])


async def send_result(session, job_data, job, is_stream=False):
    """
    Return the job results.
    """
    await _handle_result(
        session, job_data, job, JOB_DONE_URL, "Results sent.", is_stream=is_stream
    )


async def stream_result(session, job_data, job):
    """
    Return the stream job results.
    """
    await _handle_result(
        session, job_data, job, JOB_STREAM_URL, "Intermediate results sent."
    )



"""
Test cleanup.py
"""

from unittest.mock import patch

from runpod.serverless.utils import rp_cleanup


def test_clean_no_folders():
    """
    Test clean() with no folders.
    """
    with patch("shutil.rmtree") as mock_rmtree, patch(
        "os.remove"
    ) as mock_remove, patch("os.path.exists", return_value=False):
        rp_cleanup.clean()
        assert mock_rmtree.call_count == 3
        mock_rmtree.assert_any_call("input_objects", ignore_errors=True)
        mock_rmtree.assert_any_call("output_objects", ignore_errors=True)
        mock_rmtree.assert_any_call("job_files", ignore_errors=True)
        mock_remove.assert_not_called()


def test_clean_with_output_zip():
    """
    Test clean() with output.zip.
    """
    with patch("shutil.rmtree") as mock_rmtree, patch(
        "os.remove"
    ) as mock_remove, patch("os.path.exists", return_value=True):
        rp_cleanup.clean()
        assert mock_rmtree.call_count == 3
        mock_remove.assert_called_once_with("output.zip")


def test_clean_with_folders():
    """
    Test clean() with folders.
    """
    with patch("shutil.rmtree") as mock_rmtree, patch(
        "os.remove"
    ) as mock_remove, patch("os.path.exists", return_value=False):
        rp_cleanup.clean(["test_folder1", "test_folder2"])
        assert mock_rmtree.call_count == 5
        mock_rmtree.assert_any_call("test_folder1", ignore_errors=True)
        mock_rmtree.assert_any_call("test_folder2", ignore_errors=True)
        mock_remove.assert_not_called()



"""
RunPod | CLI | Exec | Functions
"""

from runpod.cli.utils import ssh_cmd


def python_over_ssh(pod_id, file):
    """
    Runs a Python file over SSH.
    """
    ssh = ssh_cmd.SSHConnection(pod_id)
    ssh.put_file(file, f"/root/{file}")
    ssh.run_commands([f"python3.10 /root/{file}"])
    ssh.close()



"""
PodWorker | modules | logging.py

Log Levels (Level - Value - Description)

NOTSET - 0 - No logging is configured, the logging system is effectively disabled.
DEBUG - 1 - Detailed information, typically of interest only when diagnosing problems. (Default)
INFO - 2 - Confirmation that things are working as expected.
WARN - 3 - An indication that something unexpected happened.
ERROR - 4 - Serious problem, the software has not been able to perform some function.
"""

import json
import os
from typing import Optional

MAX_MESSAGE_LENGTH = 4096
LOG_LEVELS = ["NOTSET", "TRACE", "DEBUG", "INFO", "WARN", "ERROR"]


def _validate_log_level(log_level):
    """
    Checks the debug level and returns the debug level name.
    """
    if isinstance(log_level, str):
        log_level = log_level.upper()

        if log_level not in LOG_LEVELS:
            raise ValueError(f"Invalid debug level: {log_level}")

        return log_level

    if isinstance(log_level, int):
        if log_level < 0 or log_level >= len(LOG_LEVELS):
            raise ValueError(f"Invalid debug level: {log_level}")

        return LOG_LEVELS[log_level]

    raise ValueError(f"Invalid debug level: {log_level}")


class RunPodLogger:
    """Singleton class for logging."""

    __instance = None
    level = _validate_log_level(
        os.environ.get(
            "RUNPOD_LOG_LEVEL", os.environ.get("RUNPOD_DEBUG_LEVEL", "DEBUG")
        )
    )

    def __new__(cls):
        if RunPodLogger.__instance is None:
            RunPodLogger.__instance = object.__new__(cls)
        return RunPodLogger.__instance

    def set_level(self, new_level):
        """
        Set the debug level for logging.
        Can be set to the name or value of the debug level.
        """
        self.level = _validate_log_level(new_level)
        self.info(f"Log level set to {self.level}")

    def log(self, message, message_level="INFO", job_id=None):
        """
        Log message to stdout if RUNPOD_DEBUG is true.
        """
        if self.level == "NOTSET":
            return

        level_index = LOG_LEVELS.index(self.level)
        if level_index > LOG_LEVELS.index(message_level) and message_level != "TIP":
            return

        message = str(message)
        # Truncate message over 10MB, remove chunk from the middle
        if len(message) > MAX_MESSAGE_LENGTH:
            half_max_length = MAX_MESSAGE_LENGTH // 2
            truncated_amount = len(message) - MAX_MESSAGE_LENGTH
            truncation_note = f"\n...TRUNCATED {truncated_amount} CHARACTERS...\n"
            message = (
                message[:half_max_length] + truncation_note + message[-half_max_length:]
            )

        if os.environ.get("RUNPOD_ENDPOINT_ID"):
            log_json = {"requestId": job_id, "message": message, "level": message_level}
            print(json.dumps(log_json), flush=True)
            return

        if job_id:
            message = f"{job_id} | {message}"

        print(f"{message_level.ljust(7)}| {message}", flush=True)
        return

    def secret(self, secret_name, secret):
        """
        Censors secrets for logging.
        Replaces everything except the first and last characters with *
        """
        secret = str(secret)
        redacted_secret = secret[0] + "*" * (len(secret) - 2) + secret[-1]
        self.info(f"{secret_name}: {redacted_secret}")

    def debug(self, message, request_id: Optional[str] = None):
        """
        debug log
        """
        self.log(message, "DEBUG", request_id)

    def info(self, message, request_id: Optional[str] = None):
        """
        info log
        """
        self.log(message, "INFO", request_id)

    def warn(self, message, request_id: Optional[str] = None):
        """
        warn log
        """
        self.log(message, "WARN", request_id)

    def error(self, message, request_id: Optional[str] = None):
        """
        error log
        """
        self.log(message, "ERROR", request_id)

    def tip(self, message):
        """
        tip log
        """
        self.log(message, "TIP")

    def trace(self, message, request_id: Optional[str] = None):
        """
        trace log (buffered until flushed)
        """
        self.log(message, "TRACE", request_id)



""" RunPod | API Wrapper | Mutations | Container Registry Auth """


def generate_container_registry_auth(name: str, username: str, password: str):
    """
    Generate a GraphQL mutation string to save container registry authentication details.

    Args:
        name (str): The name of the container registry.
        username (str): The username for authentication.
        password (str): The password for authentication.

    Returns:
        str: The GraphQL mutation string.
    """
    # Prepare the input dictionary
    input_dict = {"name": name, "username": username, "password": password}

    # Convert the input dictionary to a string, properly formatted for GraphQL
    input_str = ", ".join(f'{key}: "{value}"' for key, value in input_dict.items())

    return f"""
    mutation SaveRegistryAuth {{
        saveRegistryAuth(input: {{{input_str}}}) {{
            id
            name
        }}
    }}
    """


def update_container_registry_auth(registry_auth_id: str, username: str, password: str):
    """
    Generate a GraphQL mutation string to update registry authentication details.

    Args:
        registry_auth_id (str): The id of the container registry authentication
        username (str): The username for authentication.
        password (str): The password for authentication.

    Returns:
        str: The GraphQL mutation string.
    """
    # Prepare the input dictionary
    input_dict = {"id": registry_auth_id, "username": username, "password": password}

    # Convert the input dictionary to a string, properly formatted for GraphQL
    input_str = ", ".join(f'{key}: "{value}"' for key, value in input_dict.items())

    return f"""
    mutation UpdateRegistryAuth {{
        updateRegistryAuth(input: {{{input_str}}}) {{
            id
            name
        }}
    }}
    """


def delete_container_registry_auth(registry_auth_id: str):
    """
    Generate a GraphQL mutation string to delete registry authentication details.

    Args:
        registry_auth_id (str): The id of the container registry authentication

    Returns:
        str: The GraphQL mutation string.
    """

    return f"""
    mutation DeleteRegistryAuth {{
        deleteRegistryAuth(registryAuthId: "{registry_auth_id}")
    }}
    """



""" Unit tests for the handler module.
"""

import unittest

from runpod.serverless.modules.rp_handler import is_generator


class TestIsGenerator(unittest.TestCase):
    """Tests for the is_generator function."""

    def test_regular_function(self):
        """Test that a regular function is not a generator."""

        def regular_func():
            return "I'm a regular function!"

        self.assertFalse(is_generator(regular_func))

    def test_generator_function(self):
        """Test that a generator function is a generator."""

        def generator_func():
            yield "I'm a generator function!"

        self.assertTrue(is_generator(generator_func))

    def test_async_function(self):
        """Test that an async function is not a generator."""

        async def async_func():
            return "I'm an async function!"

        self.assertFalse(is_generator(async_func))

    def test_async_generator_function(self):
        """Test that an async generator function is a generator."""

        async def async_gen_func():
            yield "I'm an async generator function!"

        self.assertTrue(is_generator(async_gen_func))



"""
RunPod | Python | Endpoint Runner
"""

import time
from typing import Any, Dict, Optional

import requests
from requests.adapters import HTTPAdapter, Retry

from runpod.endpoint.helpers import (
    API_KEY_NOT_SET_MSG,
    FINAL_STATES,
    UNAUTHORIZED_MSG,
    is_completed,
)


# ---------------------------------------------------------------------------- #
#                                    Client                                    #
# ---------------------------------------------------------------------------- #
class RunPodClient:
    """A client for running endpoint calls."""

    def __init__(self):
        """
        Initialize a RunPodClient instance.

        Raises:
            RuntimeError: If the API key has not been initialized.
        """
        from runpod import (  # pylint: disable=import-outside-toplevel, cyclic-import
            api_key,
            endpoint_url_base,
        )

        if api_key is None:
            raise RuntimeError(API_KEY_NOT_SET_MSG)

        self.rp_session = requests.Session()
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[408, 429])
        self.rp_session.mount("http://", HTTPAdapter(max_retries=retries))

        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }

        self.endpoint_url_base = endpoint_url_base

    def _request(
        self, method: str, endpoint: str, data: Optional[dict] = None, timeout: int = 10
    ):
        """
        Make a request to the specified endpoint using the given HTTP method.

        Args:
            method: The HTTP method to use ('GET' or 'POST').
            endpoint: The endpoint path to which the request will be made.
            data: The JSON payload to send with the request.
            timeout: The number of seconds to wait for the server to send data before giving up.

        Returns:
            The JSON response from the server.

        Raises:
            RuntimeError: If the response returns a 401 Unauthorized status.
            requests.HTTPError: If the response contains an unsuccessful status code.
        """
        url = f"{self.endpoint_url_base}/{endpoint}"
        response = self.rp_session.request(
            method, url, headers=self.headers, json=data, timeout=timeout
        )

        if response.status_code == 401:
            raise RuntimeError(UNAUTHORIZED_MSG)

        response.raise_for_status()
        return response.json()

    def post(self, endpoint: str, data: dict, timeout: int = 10):
        """Post to the endpoint."""
        return self._request("POST", endpoint, data, timeout)

    def get(self, endpoint: str, timeout: int = 10):
        """Get from the endpoint."""
        return self._request("GET", endpoint, timeout=timeout)


# ---------------------------------------------------------------------------- #
#                                      Job                                     #
# ---------------------------------------------------------------------------- #
class Job:
    """Represents a job to be run on the RunPod service."""

    def __init__(self, endpoint_id: str, job_id: str, client: RunPodClient):
        """
        Initialize a Job instance with the given endpoint ID and job ID.

        Args:
            endpoint_id: The identifier for the endpoint.
            job_id: The identifier for the job.
            client: An instance of the RunPodClient to make requests with.
        """
        self.endpoint_id = endpoint_id
        self.job_id = job_id
        self.rp_client = client

        self.job_status = None
        self.job_output = None

    def _fetch_job(self, source: str = "status") -> Dict[str, Any]:
        """Returns the raw json of the status, raises an exception if invalid"""
        status_url = f"{self.endpoint_id}/{source}/{self.job_id}"
        job_state = self.rp_client.get(endpoint=status_url)

        if is_completed(job_state["status"]):
            self.job_status = job_state["status"]
            self.job_output = job_state.get("output", None)

        return job_state

    def status(self):
        """Returns the status of the job request."""
        if self.job_status is not None:
            return self.job_status

        return self._fetch_job()["status"]

    def output(self, timeout: int = 0) -> Any:
        """
        Returns the output of the job request.

        Args:
            timeout: The number of seconds to wait for the server to send data before giving up.
        """
        if timeout > 0:
            while not is_completed(self.status()):
                time.sleep(1)
                timeout -= 1
                if timeout <= 0:
                    raise TimeoutError("Job timed out.")

        if self.job_output is not None:
            return self.job_output

        return self._fetch_job().get("output", None)

    def stream(self) -> Any:
        """Returns a generator that yields the output of the job request."""
        while True:
            time.sleep(1)
            stream_partial = self._fetch_job(source="stream")
            if (
                stream_partial["status"] not in FINAL_STATES
                or len(stream_partial["stream"]) > 0
            ):
                for chunk in stream_partial.get("stream", []):
                    yield chunk["output"]
            elif stream_partial["status"] in FINAL_STATES:
                break

    def cancel(self, timeout: int = 3) -> Any:
        """
        Cancels the job and returns the result of the cancellation request.

        Args:
            timeout: The number of seconds to wait for the server to respond before giving up.
        """
        return self.rp_client.post(
            f"{self.endpoint_id}/cancel/{self.job_id}", data=None, timeout=timeout
        )


# ---------------------------------------------------------------------------- #
#                                   Endpoint                                   #
# ---------------------------------------------------------------------------- #
class Endpoint:
    """Manages an endpoint to run jobs on the RunPod service."""

    def __init__(self, endpoint_id: str):
        """
        Initialize an Endpoint instance with the given endpoint ID.

        Args:
            endpoint_id: The identifier for the endpoint.

        Example:
            >>> endpoint = runpod.Endpoint("ENDPOINT_ID")
            >>> run_request = endpoint.run({"your_model_input_key": "your_model_input_value"})
            >>> print(run_request.status())
            >>> print(run_request.output())
        """
        self.endpoint_id = endpoint_id
        self.rp_client = RunPodClient()

    def run(self, request_input: Dict[str, Any]) -> Job:
        """
        Run the endpoint with the given input.

        Args:
            request_input: The input to pass into the endpoint.

        Returns:
            A Job instance for the run request.
        """
        if not request_input.get("input"):
            request_input = {"input": request_input}

        job_request = self.rp_client.post(f"{self.endpoint_id}/run", request_input)
        return Job(self.endpoint_id, job_request["id"], self.rp_client)

    def run_sync(
        self, request_input: Dict[str, Any], timeout: int = 86400
    ) -> Dict[str, Any]:
        """
        Run the endpoint with the given input synchronously.

        Args:
            request_input: The input to pass into the endpoint.
        """
        if not request_input.get("input"):
            request_input = {"input": request_input}

        job_request = self.rp_client.post(
            f"{self.endpoint_id}/runsync", request_input, timeout=timeout
        )

        if job_request["status"] in FINAL_STATES:
            return job_request.get("output", None)

        return Job(self.endpoint_id, job_request["id"], self.rp_client).output(
            timeout=timeout
        )

    def health(self, timeout: int = 3) -> Dict[str, Any]:
        """
        Check the health of the endpoint (number/state of workers, number/state of requests).

        Args:
            timeout: The number of seconds to wait for the server to respond before giving up.
        """
        return self.rp_client.get(f"{self.endpoint_id}/health", timeout=timeout)

    def purge_queue(self, timeout: int = 3) -> Dict[str, Any]:
        """
        Purges the endpoint's job queue and returns the result of the purge request.

        Args:
            timeout: The number of seconds to wait for the server to respond before giving up.
        """
        return self.rp_client.post(
            f"{self.endpoint_id}/purge-queue", data=None, timeout=timeout
        )



""" Tests for my_module | bucket utilities """

import importlib
import io
import os
import time
import unittest
from unittest.mock import MagicMock, Mock, patch

from runpod.serverless.utils import (
    rp_upload,
    upload_file_to_bucket,
    upload_in_memory_object,
)
from runpod.serverless.utils.rp_upload import get_boto_client

BUCKET_CREDENTIALS = {
    "endpointUrl": "https://your-bucket-endpoint-url.com",
    "accessId": "your_access_key_id",
    "accessSecret": "your_secret_access_key",
}


class TestBotoConfig(unittest.TestCase):
    """Tests for boto config"""

    def setUp(self) -> None:
        self.original_environ = os.environ.copy()
        self.mock_transfer_config = MagicMock()
        self.mock_boto_client = MagicMock()

    def tearDown(self):
        os.environ = self.original_environ

    def test_get_boto_client(self):
        """
        Tests get_boto_client
        """
        # Define the bucket credentials
        bucket_creds = BUCKET_CREDENTIALS

        # Mock boto3.session.Session
        with patch("boto3.session.Session") as mock_session, patch(
            "runpod.serverless.utils.rp_upload.TransferConfig"
        ) as mock_transfer_config:
            mock_session.return_value.client.return_value = self.mock_boto_client
            mock_transfer_config.return_value = self.mock_transfer_config

            # Call get_boto_client
            boto_client, transfer_config = get_boto_client(bucket_creds)

            # Check if boto_client and transfer_config are correct
            self.assertEqual(boto_client, self.mock_boto_client)
            self.assertEqual(transfer_config, self.mock_transfer_config)

            # Check if boto3.session.Session was called with the correct arguments
            mock_session.assert_called_once_with()

            # Check if boto_client was called with the correct arguments
            mock_session.return_value.client.assert_called_once_with(
                "s3",
                endpoint_url=bucket_creds["endpointUrl"],
                aws_access_key_id=bucket_creds["accessId"],
                aws_secret_access_key=bucket_creds["accessSecret"],
                config=unittest.mock.ANY,
                region_name=None,
            )

            creds_s3 = bucket_creds.copy()
            creds_s3["endpointUrl"] = (
                "https://bucket-name.s3.region-code.amazonaws.com/key-name"
            )

            boto_client, transfer_config = get_boto_client(creds_s3)

            mock_session.return_value.client.assert_called_with(
                "s3",
                endpoint_url=creds_s3["endpointUrl"],
                aws_access_key_id=bucket_creds["accessId"],
                aws_secret_access_key=bucket_creds["accessSecret"],
                config=unittest.mock.ANY,
                region_name="region-code",
            )

            creds_do = bucket_creds.copy()
            creds_do["endpointUrl"] = (
                "https://name.region-code.digitaloceanspaces.com/key-name"
            )

            boto_client, transfer_config = get_boto_client(creds_do)

            mock_session.return_value.client.assert_called_with(
                "s3",
                endpoint_url=creds_do["endpointUrl"],
                aws_access_key_id=bucket_creds["accessId"],
                aws_secret_access_key=bucket_creds["accessSecret"],
                config=unittest.mock.ANY,
                region_name="region-code",
            )

    def test_get_boto_client_environ(self):
        """
        Tests get_boto_client with environment variables
        """
        assert rp_upload.get_boto_client()[0] is None

        os.environ["BUCKET_ENDPOINT_URL"] = "https://your-bucket-endpoint-url.com"
        os.environ["BUCKET_ACCESS_KEY_ID"] = "your_access_key_id"
        os.environ["BUCKET_SECRET_ACCESS_KEY"] = "your_secret_access_key"

        importlib.reload(rp_upload)

        with patch("boto3.session.Session") as mock_session, patch(
            "runpod.serverless.utils.rp_upload.TransferConfig"
        ) as mock_transfer_config:
            mock_session.return_value.client.return_value = self.mock_boto_client
            mock_transfer_config.return_value = self.mock_transfer_config

            boto_client, transfer_config = rp_upload.get_boto_client()

            assert boto_client == self.mock_boto_client
            assert transfer_config == self.mock_transfer_config


# ---------------------------------------------------------------------------- #
#                                 Upload Image                                 #
# ---------------------------------------------------------------------------- #


class TestUploadImage(unittest.TestCase):
    """Tests for upload_image"""

    @patch("runpod.serverless.utils.rp_upload.get_boto_client")
    @patch("builtins.open")
    @patch("runpod.serverless.utils.rp_upload.os.makedirs")
    def test_upload_image_local(self, mock_makedirs, mock_open, mock_get_boto_client):
        """
        Test upload_image function when there is no boto client
        """
        # Mocking get_boto_client to return None
        mock_get_boto_client.return_value = (None, None)

        # Mocking the context manager of Image.open
        mock_file = mock_open.return_value.__enter__.return_value
        mock_file.read.return_value = b"simulated_uploaded"
        mock_file.__exit__.return_value = False

        result = rp_upload.upload_image("job_id", "image_location")

        # Assert that image is saved locally
        assert "simulated_uploaded" in result
        mock_makedirs.assert_called_once()

    @patch("runpod.serverless.utils.rp_upload.get_boto_client")
    @patch("builtins.open")
    def test_upload_image_s3(self, mock_open, mock_get_boto_client):
        """
        Test upload_image function when there is a boto client
        """
        # Mocking boto_client
        mock_boto_client = Mock()
        mock_boto_client.put_object = Mock()
        mock_boto_client.generate_presigned_url = Mock(return_value="presigned_url")

        mock_get_boto_client.return_value = (mock_boto_client, None)

        # Mocking the context manager of Image.open
        mock_image = Mock()
        mock_image.format = "PNG"
        mock_open.return_value.__enter__.return_value = mock_image

        result = rp_upload.upload_image("job_id", "image_location")

        # Assert the image is uploaded to S3
        assert result == "presigned_url"
        mock_open.assert_called_once_with("image_location", "rb")
        mock_boto_client.put_object.assert_called_once()
        mock_boto_client.generate_presigned_url.assert_called_once()


class TestUploadUtility(unittest.TestCase):
    """Tests for upload utility"""

    @patch("runpod.serverless.utils.rp_upload.get_boto_client")
    def test_upload_file_to_bucket(self, mock_get_boto_client):
        """
        Tests upload_file_to_bucket
        """
        # Mock boto_client and transfer_config
        mock_boto_client = MagicMock()
        mock_transfer_config = MagicMock()

        mock_get_boto_client.return_value = (mock_boto_client, mock_transfer_config)

        # Define the file name and file location
        file_name = "example.txt"
        file_location = "/path/to/your/local/file/example.txt"

        # Mock os.path.getsize to return a file size
        with patch("os.path.getsize", return_value=1024):
            upload_file_to_bucket(file_name, file_location, BUCKET_CREDENTIALS)

        # Check if get_boto_client was called with the correct arguments
        mock_get_boto_client.assert_called_once_with(BUCKET_CREDENTIALS)

        # Check if upload_file was called with the correct arguments
        upload_file_args = {
            "Filename": file_location,
            "Bucket": str(time.strftime("%m-%y")),
            "Key": file_name,
            "Config": mock_transfer_config,
            "Callback": unittest.mock.ANY,
        }
        mock_boto_client.upload_file.assert_called_once_with(**upload_file_args)

        # Check if generate_presigned_url was called with the correct arguments
        mock_boto_client.generate_presigned_url.assert_called_once_with(
            "get_object",
            Params={"Bucket": str(time.strftime("%m-%y")), "Key": file_name},
            ExpiresIn=604800,
        )

    @patch("runpod.serverless.utils.rp_upload.get_boto_client")
    def test_upload_in_memory_object(self, mock_get_boto_client):
        """
        Tests upload_in_memory_object
        """
        # Mock boto_client and transfer_config
        mock_boto_client = MagicMock()
        mock_transfer_config = MagicMock()

        mock_get_boto_client.return_value = (mock_boto_client, mock_transfer_config)

        # Define the file name and file data (bytes)
        file_name = "example.txt"
        file_data = b"This is an example text."

        upload_in_memory_object(file_name, file_data, BUCKET_CREDENTIALS)

        # Check if get_boto_client was called with the correct arguments
        mock_get_boto_client.assert_called_once_with(BUCKET_CREDENTIALS)

        # Check if upload_fileobj was called with the correct arguments
        mock_boto_client.upload_fileobj.assert_called_once_with(
            unittest.mock.ANY,
            str(time.strftime("%m-%y")),
            file_name,
            Config=mock_transfer_config,
            Callback=unittest.mock.ANY,
        )

        # Check if BytesIO was called with the correct arguments
        file_data_buffer = mock_boto_client.upload_fileobj.call_args[0][0]
        self.assertIsInstance(file_data_buffer, io.BytesIO)

        # Check if generate_presigned_url was called with the correct arguments
        mock_boto_client.generate_presigned_url.assert_called_once_with(
            "get_object",
            Params={"Bucket": str(time.strftime("%m-%y")), "Key": file_name},
            ExpiresIn=604800,
        )



""" Test suite for the generate_container_registry_auth function. """

import unittest

from runpod.api.mutations.container_register_auth import (
    delete_container_registry_auth,
    generate_container_registry_auth,
    update_container_registry_auth
)


class TestGenerateContainerRegistryAuth(unittest.TestCase):
    """Test suite for the generate_container_registry_auth function."""

    def test_generate_container_registry_auth(self):
        """
        Test that the generate_container_registry_auth function produces the correct
        GraphQL mutation string with the provided name, username, and password.
        """
        # Define test inputs
        name = "testRegistry"
        username = "testUser"
        password = "testPass"

        # Generate the actual mutation string
        actual_mutation = generate_container_registry_auth(
            name, username, password
        ).strip()

        # Verify key components of the mutation string
        self.assertIn("mutation SaveRegistryAuth", actual_mutation)
        self.assertIn(
            'saveRegistryAuth(input: {name: "testRegistry", username: "testUser", password: "testPass"})',  # pylint: disable=line-too-long
            actual_mutation,
        )
        self.assertIn("id", actual_mutation)
        self.assertIn("name", actual_mutation)

    def test_update_container_registry_auth(self):
        """
        Test that the update_container_registry_auth function produces the correct
        GraphQL mutation string with the provided registry_auth_id, username and password.
        """
        # Define test inputs
        registry_auth_id = "testAuthId"
        username = "testUser"
        password = "testPass"

        # Generate the actual mutation string
        actual_mutation = update_container_registry_auth(
            registry_auth_id, username, password
        ).strip()

        # Verify key components of the mutation string
        self.assertIn("mutation UpdateRegistryAuth", actual_mutation)
        self.assertIn(
            'updateRegistryAuth(input: {id: "testAuthId", username: "testUser", password: "testPass"})',  # pylint: disable=line-too-long
            actual_mutation,
        )
        self.assertIn("id", actual_mutation)
        self.assertIn("name", actual_mutation)

    def test_delete_container_registry_auth(self):
        """
        Test that the delete_container_registry_auth function produces the correct
        GraphQL mutation string with the provided registry_auth_id
        """
        # Define test inputs
        registry_auth_id = "testAuthId"

        # Generate the actual mutation string
        actual_mutation = delete_container_registry_auth(
            registry_auth_id
        ).strip()

        # Verify key components of the mutation string
        self.assertIn("mutation DeleteRegistryAuth", actual_mutation)
        self.assertIn(
            'deleteRegistryAuth(registryAuthId: "testAuthId")',  # pylint: disable=line-too-long
            actual_mutation,
        )



"""
Example of calling an endpoint using the RunPod Python Language Library.
"""

import runpod

# Set your global API key with `runpod config` or uncomment the line below:
# runpod.api_key = "YOUR_RUNPOD_API_KEY"

endpoint = runpod.Endpoint("sdxl")  # Where "sdxl" is the endpoint ID

try:
    # Run the endpoint synchronously, blocking until the endpoint run is complete.
    run_request = endpoint.run_sync(
        {"input": {"prompt": "a photo of a horse the size of a Boeing 787"}},
        timeout=60,  # Seconds
    )

    print(run_request)
except TimeoutError as err:
    print("Job timed out.")



""" A template for a Llama2 handler file. """

# pylint: skip-file

import inspect

from transformers import HfApi

import runpod

SELECTED_MODEL = "<<MODEL_NAME>>"


def get_model_framework(model_name):
    api = HfApi()
    model_files = api.model_info(model_name).files

    # Check the files associated with the model
    if "pytorch_model.bin" in model_files:
        return "PyTorch"
    elif "tf_model.h5" in model_files:
        return "TensorFlow"
    else:
        return "Unknown"


def prepare_inputs(text, **kwargs):
    # Filter kwargs based on what the tokenizer accepts
    filtered_args = {k: v for k, v in kwargs.items() if k in valid_args}

    inputs = tokenizer(text, return_tensors="pt", **filtered_args)
    return inputs


def handle_request(text, **input_args):
    inputs = prepare_inputs(text, **input_args)
    with torch.no_grad():
        outputs = model(**inputs)
    return process_outputs(outputs)


runpod.serverless.start({"handler": handle_request})



"""
RunPod | API Wrapper | Queries | GPUs
"""

QUERY_POD = """
query myPods {
    myself {
        pods {
            id
            containerDiskInGb
            costPerHr
            desiredStatus
            dockerArgs
            dockerId
            env
            gpuCount
            imageName
            lastStatusChange
            machineId
            memoryInGb
            name
            podType
            port
            ports
            uptimeSeconds
            vcpuCount
            volumeInGb
            volumeMountPath
            runtime {
                ports{
                    ip
                    isIpPublic
                    privatePort
                    publicPort
                    type
                }
            }
            machine {
                gpuDisplayName
            }
        }
    }
}
"""


def generate_pod_query(pod_id):
    """
    Generate a query for a specific GPU type
    """

    return f"""
    query pod {{
        pod(input: {{podId: "{pod_id}"}}) {{
            id
            containerDiskInGb
            costPerHr
            desiredStatus
            dockerArgs
            dockerId
            env
            gpuCount
            imageName
            lastStatusChange
            machineId
            memoryInGb
            name
            podType
            port
            ports
            uptimeSeconds
            vcpuCount
            volumeInGb
            volumeMountPath
            runtime {{
                ports {{
                    ip
                    isIpPublic
                    privatePort
                    publicPort
                    type
                }}
            }}
            machine {{
                gpuDisplayName
            }}
        }}
    }}
    """



""" Helper functions for the RunPod Endpoint API. """

FINAL_STATES = ["COMPLETED", "FAILED", "TIMED_OUT"]

# Exception Messages
UNAUTHORIZED_MSG = "401 Unauthorized | Make sure Runpod API key is set and valid."
API_KEY_NOT_SET_MSG = (
    "Expected `run_pod.api_key` to be initialized. "
    "You can solve this by setting `run_pod.api_key = 'your-key'. "
    "An API key can be generated at "
    "https://runpod.io/console/user/settings"
)


def is_completed(status: str) -> bool:
    """Returns true if status is one of the possible final states for a serverless request."""
    return status in ["COMPLETED", "FAILED", "TIMED_OUT", "CANCELLED"]



"""
Test rp_http.py module.
"""

# pylint: disable=too-few-public-methods

import gc
import json
import unittest
from unittest.mock import AsyncMock, patch

import aiohttp

from runpod.serverless.modules import rp_http


class MockRequestInfo:
    """Mock aiohttp.RequestInfo class."""

    def __init__(self, *args, **kwargs):
        del args, kwargs
        self.url = "http://test_url"
        self.method = "POST"
        self.headers = {"Content-Type": "application/json"}
        self.real_url = "http://test_url"

    real_url = "http://test_url"


class TestHTTP(unittest.IsolatedAsyncioTestCase):
    """Test HTTP module."""

    async def asyncSetUp(self) -> None:
        self.job = {"id": "test_id"}
        self.job_data = {"output": "test_output"}

    async def asyncTearDown(self) -> None:
        gc.collect()

    async def test_send_result(self):
        """
        Test send_result function.
        """
        with patch("runpod.serverless.modules.rp_http.log") as mock_log, patch(
            "runpod.serverless.modules.rp_http.RetryClient"
        ) as mock_retry:

            mock_retry.return_value.post.return_value = AsyncMock()
            mock_retry.return_value.post.return_value.__aenter__.return_value.text.return_value = (
                "response text"  # pylint: disable=line-too-long
            )

            send_return_local = await rp_http.send_result(
                AsyncMock(), self.job_data, self.job
            )

            assert send_return_local is None
            assert mock_log.debug.call_count == 1
            assert mock_log.error.call_count == 0
            assert mock_log.info.call_count == 1

            mock_retry.return_value.post.assert_called_with(
                "JOB_DONE_URL" + "&isStream=false",
                data=str(json.dumps(self.job_data, ensure_ascii=False)),
                headers={
                    "charset": "utf-8",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
                raise_for_status=True,
            )

    async def test_send_result_client_response_error(self):
        """
        Test send_result function with ClientResponseError.
        """

        def mock_request_info_init(self, *args, **kwargs):
            """
            Mock aiohttp.RequestInfo.__init__ method.
            """
            del args, kwargs
            self.url = "http://test_url"
            self.method = "POST"
            self.headers = {"Content-Type": "application/json"}
            self.real_url = "http://test_url"

        with patch("runpod.serverless.modules.rp_http.log") as mock_log, patch(
            "runpod.serverless.modules.rp_http.RetryClient"
        ) as mock_retry, patch.object(
            aiohttp.RequestInfo, "__init__", mock_request_info_init
        ):

            mock_retry.side_effect = aiohttp.ClientResponseError(
                request_info=MockRequestInfo,
                history=None,
                status=500,
                message="Error message",
            )

            send_return_local = await rp_http.send_result(
                AsyncMock(), self.job_data, self.job
            )

            assert send_return_local is None
            assert mock_log.debug.call_count == 0
            assert mock_log.error.call_count == 1
            assert mock_log.info.call_count == 1

    async def test_send_result_type_error(self):
        """
        Test send_result function with TypeError.
        """
        with patch("runpod.serverless.modules.rp_http.log") as mock_log, patch(
            "runpod.serverless.modules.rp_http.json.dumps"
        ) as mock_dumps, patch(
            "runpod.serverless.modules.rp_http.RetryClient"
        ) as mock_retry:

            mock_dumps.side_effect = TypeError("Forced exception")

            send_return_local = await rp_http.send_result(
                aiohttp.ClientSession(), self.job_data, self.job
            )

            assert send_return_local is None
            assert mock_log.debug.call_count == 0
            assert mock_log.error.call_count == 1
            assert mock_log.info.call_count == 1
            assert mock_retry.return_value.post.call_count == 0
            mock_log.error.assert_called_with(
                "Error while returning job result. | Forced exception", "test_id"
            )  # pylint: disable=line-too-long

    async def test_stream_result(self):
        """
        Test stream_result function.
        """
        with patch("runpod.serverless.modules.rp_http.log") as mock_log, patch(
            "runpod.serverless.modules.rp_http.RetryClient"
        ) as mock_retry:

            mock_retry.return_value.post.return_value = AsyncMock()
            mock_retry.return_value.post.return_value.__aenter__.return_value.text.return_value = (
                "response text"  # pylint: disable=line-too-long
            )

            send_return_local = await rp_http.stream_result(
                AsyncMock(), self.job_data, self.job
            )

            assert send_return_local is None
            assert mock_log.debug.call_count == 1
            assert mock_log.error.call_count == 0
            assert mock_log.info.call_count == 0

            mock_retry.return_value.post.assert_called_with(
                "JOB_STREAM_URL" + "&isStream=false",
                data=str(json.dumps(self.job_data, ensure_ascii=False)),
                headers={
                    "charset": "utf-8",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
                raise_for_status=True,
            )


if __name__ == "__main__":
    unittest.main()



"""
RunPod | API Wrapper | GraphQL
"""

import json
import os
from typing import Any, Dict

import requests

from runpod import error
from runpod.user_agent import USER_AGENT

HTTP_STATUS_UNAUTHORIZED = 401


def run_graphql_query(query: str) -> Dict[str, Any]:
    """
    Run a GraphQL query
    """
    from runpod import api_key  # pylint: disable=import-outside-toplevel, cyclic-import

    api_url_base = os.environ.get("RUNPOD_API_BASE_URL", "https://api.runpod.io")
    url = f"{api_url_base}/graphql"

    headers = {
        "Content-Type": "application/json",
        "User-Agent": USER_AGENT,
        "Authorization": f"Bearer {api_key}",
    }

    data = json.dumps({"query": query})
    response = requests.post(url, headers=headers, data=data, timeout=30)

    if response.status_code == HTTP_STATUS_UNAUTHORIZED:
        raise error.AuthenticationError(
            "Unauthorized request, please check your API key."
        )

    if "errors" in response.json():
        raise error.QueryError(response.json()["errors"][0]["message"], query)

    return response.json()



""" Tests for runpod.serverless.modules.rp_ping """

import importlib
import os
import unittest
from unittest.mock import patch, MagicMock

import requests

from runpod.serverless.modules import rp_ping
from runpod.serverless.modules.rp_ping import Heartbeat
from runpod.serverless.modules.worker_state import JobsProgress


class MockResponse:
    """Mock response for aiohttp"""
    url = ""
    status_code = 200


def mock_get(*args, **kwargs):
    """
    Mock get function for aiohttp
    """
    return MockResponse()


class TestPing(unittest.TestCase):
    """Tests for rp_ping"""

    def test_default_variables(self):
        """
        Tests that the variables are set with default values
        """
        heartbeat = Heartbeat()
        assert heartbeat.PING_URL == "PING_NOT_SET"
        assert heartbeat.PING_INTERVAL == 10

    @patch.dict(os.environ, {"RUNPOD_WEBHOOK_PING": "https://test.com/ping"})
    @patch.dict(os.environ, {"RUNPOD_PING_INTERVAL": "1000"})
    def test_variables(self):
        """
        Tests that the variables are set correctly
        """
        importlib.reload(rp_ping)

        heartbeat = Heartbeat()
        assert heartbeat.PING_URL == "https://test.com/ping"
        assert heartbeat.PING_INTERVAL == 1

    @patch.dict(os.environ, {"RUNPOD_PING_INTERVAL": "1000"})
    @patch(
        "runpod.serverless.modules.rp_ping.SyncClientSession.get", side_effect=mock_get
    )
    def test_start_ping(self, mock_get_return):
        """
        Tests that the start_ping function works correctly
        """
        # No RUNPOD_AI_API_KEY case
        with patch("threading.Thread.start") as mock_thread_start:
            rp_ping.Heartbeat().start_ping(test=True)
            assert mock_thread_start.call_count == 0

        os.environ["RUNPOD_AI_API_KEY"] = "test_key"

        # No RUNPOD_POD_ID case
        with patch("threading.Thread.start") as mock_thread_start:
            rp_ping.Heartbeat().start_ping(test=True)
            assert mock_thread_start.call_count == 0

        os.environ["RUNPOD_POD_ID"] = "test_pod_id"

        # No RUNPOD_WEBHOOK_PING case
        with patch("threading.Thread.start") as mock_thread_start:
            rp_ping.Heartbeat().start_ping(test=True)
            assert mock_thread_start.call_count == 0

        os.environ["RUNPOD_WEBHOOK_PING"] = "https://test.com/ping"

        importlib.reload(rp_ping)

        # Success case
        with patch("threading.Thread.start") as mock_thread_start:
            rp_ping.Heartbeat().start_ping(test=True)
            assert mock_thread_start.call_count == 1

        rp_ping.Heartbeat.PING_URL = "https://test.com/ping"
        rp_ping.Heartbeat().ping_loop(test=True)

        self.assertEqual(rp_ping.Heartbeat.PING_URL, "https://test.com/ping")

        # Exception case
        mock_get_return.side_effect = requests.RequestException("Test Error")

        with patch("runpod.serverless.modules.rp_ping.log.error") as mock_log_error:
            rp_ping.Heartbeat().ping_loop(test=True)
            assert mock_log_error.call_count == 1


@patch.dict(os.environ, {"RUNPOD_PING_INTERVAL": "1000"})
class TestHeartbeat(unittest.IsolatedAsyncioTestCase):

    @patch.dict(os.environ, {"RUNPOD_AI_API_KEY": ""})
    @patch("runpod.serverless.modules.rp_ping.log")
    def test_start_ping_no_api_key(self, mock_logger):
        """Test start_ping method when RUNPOD_AI_API_KEY is missing."""
        heartbeat = Heartbeat()
        heartbeat.start_ping()
        mock_logger.debug.assert_called_once_with(
            "Not deployed on RunPod serverless, pings will not be sent."
        )

    @patch.dict(os.environ, {"RUNPOD_POD_ID": ""})
    @patch("runpod.serverless.modules.rp_ping.log")
    def _test_start_ping_no_pod_id(self, mock_logger):
        """Test start_ping method when RUNPOD_POD_ID is missing."""
        heartbeat = Heartbeat()
        heartbeat.start_ping()
        mock_logger.info.assert_called_once_with(
            "Not running on RunPod, pings will not be sent."
        )

    @patch("runpod.serverless.modules.rp_ping.Heartbeat._send_ping")
    def test_ping_loop(self, mock_send_ping):
        """Test ping_loop runs and exits correctly in test mode."""
        heartbeat = rp_ping.Heartbeat()
        heartbeat.ping_loop(test=True)
        mock_send_ping.assert_called_once()

    @patch("runpod.serverless.modules.rp_ping.SyncClientSession.get")
    async def test_send_ping(self, mock_get):
        """Test _send_ping method sends the correct request."""
        mock_response = MagicMock()
        mock_response.url = "http://localhost/ping"
        mock_response.status_code = 200
        mock_get.return_value = mock_response

        jobs = JobsProgress()
        jobs.add("job1")
        jobs.add("job2")

        heartbeat = Heartbeat()
        heartbeat._send_ping()

        mock_get.assert_called_once()

        # Extract the arguments passed to the mock_get call
        _, kwargs = mock_get.call_args

        # Check that job_id is correct in params, ignoring other params
        assert 'params' in kwargs
        assert 'job_id' in kwargs['params']
        assert kwargs['params']['job_id'] in ["job1,job2", "job2,job1"]

    @patch("runpod.serverless.modules.rp_ping.log")
    def test_send_ping_exception(self, mock_logger):
        """Test _send_ping logs an error on exception."""
        heartbeat = Heartbeat()

        with patch.object(
            heartbeat._session,
            "get",
            side_effect=requests.RequestException("Error"),
        ):
            heartbeat._send_ping()

            mock_logger.error.assert_called_once_with(
                "Ping Request Error: Error, attempting to restart ping."
            )



""" Example of using the RunPodLogger class. """ ""

import runpod

JOB_ID = "1234567890"
log = runpod.RunPodLogger()


log.debug("A debug message")
log.info("An info message")
log.warn("A warning message")
log.error("An error message")

# Output:
# DEBUG   | A debug message
# INFO    | An info message
# WARN    | A warning message
# ERROR   | An error message


log.debug("A debug message", request_id=JOB_ID)
log.info("An info message", request_id=JOB_ID)
log.warn("A warning message", request_id=JOB_ID)
log.error("An error message", request_id=JOB_ID)

# Output:
# {"requestId": "1234567890", "message": "A debug message", "level": "DEBUG"}
# {"requestId": "1234567890", "message": "An info message", "level": "INFO"}
# {"requestId": "1234567890", "message": "A warning message", "level": "WARN"}
# {"requestId": "1234567890", "message": "An error message", "level": "ERROR"}



"""
runpod | serverless | worker_loop.py
Called to convert a container into a worker pod for the runpod serverless platform.
"""

import asyncio
import os
from typing import Any, Dict

from runpod.serverless.modules import rp_logger, rp_local, rp_ping, rp_scale

log = rp_logger.RunPodLogger()
heartbeat = rp_ping.Heartbeat()


def _is_local(config) -> bool:
    """Returns True if the worker is running locally, False otherwise."""
    if config["rp_args"].get("test_input", None):
        return True

    if os.environ.get("RUNPOD_WEBHOOK_GET_JOB", None) is None:
        return True

    return False


# ------------------------- Main Worker Running Loop ------------------------- #
def run_worker(config: Dict[str, Any]) -> None:
    """
    Starts the worker loop for multi-processing.

    This function is called when the worker is running on RunPod. This function
    starts a loop that runs indefinitely until the worker is killed.

    Args:
        config (Dict[str, Any]): Configuration parameters for the worker.
    """
    # Start pinging RunPod to show that the worker is alive.
    heartbeat.start_ping()

    # Create a JobScaler responsible for adjusting the concurrency
    job_scaler = rp_scale.JobScaler(config)
    job_scaler.start()


def main(config: Dict[str, Any]) -> None:
    """
    Checks if the worker is running locally or on RunPod.
    If running locally, the test job is run and the worker exits.
    If running on RunPod, the worker loop is created.
    """
    if _is_local(config):
        asyncio.run(rp_local.run_local(config))

    else:
        run_worker(config)



import sys
import traceback
from unittest import TestCase
from unittest.mock import patch

from runpod.serverless.modules.rp_scale import _handle_uncaught_exception


class TestHandleUncaughtException(TestCase):
    def setUp(self):
        sys.excepthook = sys.__excepthook__

    @patch("runpod.serverless.modules.rp_scale.log")
    def test_handle_uncaught_exception(self, mock_logger):
        exc_type = ValueError
        exc_value = ValueError("This is a test error")
        exc_traceback = None  # No traceback for simplicity

        _handle_uncaught_exception(exc_type, exc_value, exc_traceback)

        formatted_exception = traceback.format_exception(exc_type, exc_value, exc_traceback)

        mock_logger.error.assert_called_once()
        log_message = mock_logger.error.call_args[0][0]
        assert "Uncaught exception | " in log_message
        assert str(formatted_exception) in log_message

    @patch("runpod.serverless.modules.rp_scale.log")
    def test_handle_uncaught_exception_with_traceback(self, mock_logger):
        try:
            raise RuntimeError("This is a runtime error")
        except RuntimeError:
            exc_type, exc_value, exc_traceback = sys.exc_info()

            _handle_uncaught_exception(exc_type, exc_value, exc_traceback)

            formatted_exception = traceback.format_exception(exc_type, exc_value, exc_traceback)

            mock_logger.error.assert_called_once()
            log_message = mock_logger.error.call_args[0][0]
            assert "Uncaught exception | " in log_message
            assert str(formatted_exception) in log_message

    @patch("runpod.serverless.modules.rp_scale.log")
    def test_handle_uncaught_exception_with_no_exception(self, mock_logger):
        _handle_uncaught_exception(None, None, None)

        mock_logger.error.assert_called_once()
        log_message = mock_logger.error.call_args[0][0]
        assert "Uncaught exception | " in log_message

    def test_excepthook_not_set_when_start_not_invoked(self):
        assert sys.excepthook == sys.__excepthook__
        assert sys.excepthook != _handle_uncaught_exception



"""
Test shared functions related to authentication
"""

import importlib
import unittest
from unittest.mock import mock_open, patch


class TestAPIKey(unittest.TestCase):
    """Test the API key"""

    # The mocked TOML file
    CREDENTIALS = b"""
    [default]
    api_key = "RUNPOD_API_KEY"
    """

    @patch("builtins.open", new_callable=mock_open, read_data=CREDENTIALS)
    def test_use_file_credentials(self, mock_file):
        """
        Test that the API key is read from the credentials file
        """
        import runpod  # pylint: disable=import-outside-toplevel

        importlib.reload(runpod)
        self.assertEqual(runpod.api_key, "RUNPOD_API_KEY")
        assert mock_file.called



""" Example of creating an endpoint with the Runpod API. """

import runpod

# Set your global API key with `runpod config` or uncomment the line below:
# runpod.api_key = "YOUR_RUNPOD_API_KEY"

try:

    new_template = runpod.create_template(
        name="test", image_name="runpod/base:0.4.4", is_serverless=True
    )

    print(new_template)

    new_endpoint = runpod.create_endpoint(
        name="test",
        template_id=new_template["id"],
        gpu_ids="AMPERE_16",
        workers_min=0,
        workers_max=1,
        flashboot=True,
    )

    print(new_endpoint)

except runpod.error.QueryError as err:
    print(err)
    print(err.query)



"""
RunPod | CLI | SSH | Functions
"""

import base64
import hashlib
import os

import paramiko

from runpod.api.ctl_commands import get_user, update_user_settings

SSH_FILES = os.path.expanduser("~/.runpod/ssh")


def get_ssh_key_fingerprint(public_key):
    """
    Get the fingerprint of an SSH key
    """
    parts = public_key.split()
    if len(parts) < 2:
        raise ValueError("Invalid SSH public key")

    key_data = base64.b64decode(parts[1])
    fingerprint = hashlib.sha256(key_data).digest()
    return "SHA256:" + base64.b64encode(fingerprint).decode("utf-8").strip("=")


def get_user_pub_keys():
    """
    Get the current user's SSH keys
    """
    user = get_user()
    keys = "" if user["pubKey"] is None else user["pubKey"]

    # Parse the keys
    keys = keys.split("\n")
    keys = [key for key in keys if key != ""]

    key_list = []
    for key in keys:
        key_parts = key.split(" ")

        # Basic validation
        if len(key_parts) < 2:
            continue

        key_dict = {}
        key_dict["type"] = key_parts[0]
        key_dict["key"] = key_parts[1]
        key_dict["fingerprint"] = get_ssh_key_fingerprint(key)
        key_dict["name"] = key_parts[2] if len(key_parts) > 2 else "N/A"
        key_list.append(key_dict)

    return key_list


def generate_ssh_key_pair(filename):
    """
    Generate an RSA SSH key pair and save it to disk.

    Args:
    - filename (str):   The base filename to save the key pair.
                        The public key will have '.pub' appended to it.
    """
    os.makedirs(os.path.join(SSH_FILES), exist_ok=True)

    # Generate private key
    private_key = paramiko.RSAKey.generate(bits=2048)
    private_key.write_private_key_file(os.path.join(SSH_FILES, filename))

    # Set permissions
    os.chmod(os.path.join(SSH_FILES, filename), 0o600)

    # Generate public key
    with open(f"{SSH_FILES}/{filename}.pub", "w", encoding="UTF-8") as public_file:
        public_key = f"{private_key.get_name()} {private_key.get_base64()} {filename}"
        public_file.write(public_key)

    add_ssh_key(public_key)

    return private_key, public_key


def add_ssh_key(public_key):
    """Add an SSH public key to the current user's RunPod account.
    Checks if the key already exists before adding it.
    """
    user = get_user()
    current_keys = "" if user["pubKey"] is None else user["pubKey"]

    # Check if the key already exists
    if public_key in current_keys:
        print("Key already exists")
        return

    updated_keys = current_keys + ("\n\n" if current_keys else "") + public_key

    # Update the user's keys
    update_user_settings(f"{updated_keys}")



""" Test functions in runpod.cli.groups.project.functions module. """

import os
import shutil
import unittest
from unittest.mock import mock_open, patch

from runpod.cli.groups.project.functions import (
    STARTER_TEMPLATES,
    create_new_project,
    create_project_endpoint,
    start_project,
)


class TestCreateNewProject(unittest.TestCase):
    """Test the create_new_project function."""

    def tearDown(self):
        toml_file_location = os.path.join(os.getcwd(), "test_project")
        if os.path.exists(toml_file_location):
            shutil.rmtree(toml_file_location)

    @patch("os.makedirs")
    @patch("os.path.exists", return_value=False)
    @patch("os.getcwd", return_value="/current/path")
    @patch("runpod.cli.groups.project.functions.copy_template_files")
    def test_create_project_folder(
        self, mock_copy_template_files, mock_getcwd, mock_exists, mock_makedirs
    ):  # pylint: disable=line-too-long
        """Test that a new project folder is created if init_current_dir is False."""
        with patch("builtins.open", new_callable=mock_open):
            create_new_project("test_project", "volume_id", "11.1.1", "3.8")
        mock_makedirs.assert_called_once_with("/current/path/test_project")
        assert mock_copy_template_files.called
        assert mock_getcwd.called
        assert mock_exists.called

    @patch("os.makedirs")
    @patch("os.path.exists", return_value=False)
    @patch("os.getcwd", return_value="/tmp/testdir")
    @patch("builtins.open", new_callable=mock_open)
    def test_create_new_project_init_current_dir(
        self, mock_file_open, mock_getcwd, mock_path_exists, mock_makedirs
    ):  # pylint: disable=line-too-long
        """Test that a new project folder is not created if init_current_dir is True."""
        project_name = "test_project"
        runpod_volume_id = "12345"
        cuda_version = "11.1.1"
        python_version = "3.9"

        create_new_project(
            project_name,
            runpod_volume_id,
            cuda_version,
            python_version,
            init_current_dir=True,
        )
        mock_makedirs.assert_not_called()
        mock_file_open.assert_called_with(
            "/tmp/testdir/runpod.toml", "w", encoding="UTF-8"
        )
        assert mock_getcwd.called
        assert mock_path_exists.called is False

    @patch("os.makedirs")
    @patch("os.path.exists", return_value=False)
    @patch("os.getcwd", return_value="/current/path")
    @patch("runpod.cli.groups.project.functions.copy_template_files")
    def test_copy_template_files(
        self, mock_copy_template_files, mock_getcwd, mock_exists, mock_makedirs
    ):  # pylint: disable=line-too-long
        """Test that template files are copied to the new project folder."""
        with patch("builtins.open", new_callable=mock_open):
            create_new_project("test_project", "volume_id", "11.1.1", "3.8")
        mock_copy_template_files.assert_called_once_with(
            STARTER_TEMPLATES + "/default", "/current/path/test_project"
        )  # pylint: disable=line-too-long
        assert mock_getcwd.called
        assert mock_exists.called
        assert mock_makedirs.called

    @patch("os.path.exists", return_value=True)
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="data with <<MODEL_NAME>> placeholder",
    )  # pylint: disable=line-too-long
    def test_replace_placeholders_in_handler(
        self, mock_open_file, mock_exists
    ):  # pylint: disable=line-too-long
        """Test that placeholders in handler.py are replaced if model_name is given."""
        with patch("runpod.cli.groups.project.functions.copy_template_files"):
            create_new_project(
                "test_project", "volume_id", "11.8.0", "3.8", model_name="my_model"
            )
        assert mock_open_file.called
        assert mock_exists.called

    @patch("os.path.exists", return_value=False)
    @patch("builtins.open", new_callable=mock_open)
    def test_create_runpod_toml(self, mock_open_file, mock_exists):
        """Test that runpod.toml configuration file is created."""
        with patch("runpod.cli.groups.project.functions.copy_template_files"):
            create_new_project("test_project", "volume_id", "11.8.0", "3.8")
        toml_file_location = os.path.join(os.getcwd(), "test_project", "runpod.toml")
        mock_open_file.assert_called_with(
            toml_file_location, "w", encoding="UTF-8"
        )  # pylint: disable=line-too-long
        assert mock_exists.called

    @patch("os.path.exists", return_value=True)
    @patch("builtins.open", new_callable=mock_open, read_data="<<RUNPOD>> placeholder")
    def test_update_requirements_file(self, mock_open_file, mock_exists):
        """Test that placeholders in requirements.txt are replaced correctly."""
        with patch("runpod.cli.groups.project.functions.__version__", "dev"), patch(
            "runpod.cli.groups.project.functions.copy_template_files"
        ):
            create_new_project("test_project", "volume_id", "11.8.0", "3.8")
        assert mock_open_file.called
        assert mock_exists.called

    @patch("os.path.exists", return_value=True)
    @patch("builtins.open", new_callable=mock_open, read_data="<<RUNPOD>> placeholder")
    def test_update_requirements_file_non_dev(self, mock_open_file, mock_exists):
        """Test that placeholders in requirements.txt are replaced for non-dev versions."""
        with patch("runpod.cli.groups.project.functions.__version__", "1.0.0"), patch(
            "runpod.cli.groups.project.functions.copy_template_files"
        ):
            create_new_project("test_project", "volume_id", "11.8.0", "3.8")
        assert mock_open_file.called
        assert mock_exists.called


class TestStartProject(unittest.TestCase):
    """Test the start_project function."""

    @patch("runpod.cli.groups.project.functions.load_project_config")
    @patch("runpod.cli.groups.project.functions.get_project_pod")
    @patch("runpod.cli.groups.project.functions.attempt_pod_launch")
    @patch("runpod.cli.groups.project.functions.get_pod")
    @patch("runpod.cli.groups.project.functions.SSHConnection")
    @patch("os.getcwd", return_value="/current/path")
    def test_start_nonexistent_successfully(
        self,
        mock_getcwd,
        mock_ssh_connection,
        mock_get_pod,
        mock_attempt_pod_launch,
        mock_get_project_pod,
        mock_load_project_config,
    ):  # pylint: disable=line-too-long, too-many-arguments
        """Test that a project is launched successfully."""
        mock_load_project_config.return_value = {
            "project": {
                "uuid": "123456",
                "name": "test_project",
                "volume_mount_path": "/mount/path",
                "env_vars": {"ENV_VAR": "value"},
                "gpu": "NVIDIA GPU",
            },
            "runtime": {
                "python_version": "3.8",
                "handler_path": "handler.py",
                "requirements_path": "requirements.txt",
            },
        }

        mock_get_project_pod.return_value = False

        mock_attempt_pod_launch.return_value = {
            "id": "new_pod_id",
            "desiredStatus": "PENDING",
            "runtime": None,
        }

        mock_get_pod.return_value = {
            "id": "new_pod_id",
            "desiredStatus": "RUNNING",
            "runtime": "ONLINE",
        }

        mock_ssh_instance = mock_ssh_connection.return_value
        mock_ssh_instance.__enter__.return_value = mock_ssh_instance
        mock_ssh_instance.run_commands.return_value = None

        with patch(
            "runpod.cli.groups.project.functions.sync_directory"
        ) as mock_sync_directory:
            start_project()

        mock_attempt_pod_launch.assert_called()
        mock_get_pod.assert_called_with("new_pod_id")
        mock_ssh_connection.assert_called_with("new_pod_id")
        mock_ssh_instance.run_commands.assert_called()
        assert mock_getcwd.called
        assert mock_sync_directory.called

    @patch("runpod.cli.groups.project.functions.get_project_pod")
    @patch("runpod.cli.groups.project.functions.attempt_pod_launch")
    def test_failed_pod_launch(self, mock_attempt_pod, mock_get_pod):
        """Test that a project is not launched if pod launch fails."""
        mock_attempt_pod.return_value = None
        mock_get_pod.return_value = None

        with patch("builtins.print") as mock_print, patch(
            "runpod.cli.groups.project.functions.load_project_config"
        ):

            start_project()
            mock_print.assert_called_with(
                "Selected GPU types unavailable, try again later or use a different type."
            )  # pylint: disable=line-too-long


class TestStartProjectAPI(unittest.TestCase):
    """Test the start_project_api function."""

    @patch("runpod.cli.groups.project.functions.load_project_config")
    @patch("runpod.cli.groups.project.functions.get_project_pod")
    @patch("runpod.cli.groups.project.functions.SSHConnection")
    @patch("os.getcwd", return_value="/current/path")
    @patch("runpod.cli.groups.project.functions.sync_directory")
    def test_start_project_api_successfully(
        self,
        mock_sync_directory,
        mock_getcwd,
        mock_ssh_connection,
        mock_get_project_pod,
        mock_load_project_config,
    ):  # pylint: disable=line-too-long, too-many-arguments
        """Test that a project API is started successfully."""
        mock_load_project_config.return_value = {
            "project": {
                "uuid": "123456",
                "name": "test_project",
                "volume_mount_path": "/mount/path",
            },
            "runtime": {
                "python_version": "3.8",
                "handler_path": "handler.py",
                "requirements_path": "requirements.txt",
            },
        }

        mock_get_project_pod.return_value = {"id": "pod_id"}

        mock_ssh_instance = mock_ssh_connection.return_value
        mock_ssh_instance.__enter__.return_value = mock_ssh_instance
        mock_ssh_instance.run_commands.return_value = None

        start_project()

        mock_get_project_pod.assert_called_with("123456")
        mock_ssh_connection.assert_called_with({"id": "pod_id"})
        mock_sync_directory.assert_called_with(
            mock_ssh_instance, "/current/path", "/mount/path/123456/dev"
        )
        mock_ssh_instance.run_commands.assert_called()
        assert mock_getcwd.called


class TestCreateProjectEndpoint(unittest.TestCase):
    """Test the create_project_endpoint function."""

    @patch("runpod.cli.groups.project.functions.SSHConnection")
    @patch("runpod.cli.groups.project.functions.load_project_config")
    @patch("runpod.cli.groups.project.functions.create_template")
    @patch("runpod.cli.groups.project.functions.create_endpoint")
    @patch("runpod.cli.groups.project.functions.update_endpoint_template")
    @patch("runpod.cli.groups.project.functions.get_project_pod")
    @patch("runpod.cli.groups.project.functions.get_project_endpoint")
    def test_create_project_endpoint(
        self,
        mock_get_project_endpoint,
        mock_get_project_pod,
        mock_update_endpoint,
        mock_create_endpoint,  # pylint: disable=too-many-arguments,line-too-long
        mock_create_template,
        mock_load_project_config,
        mock_ssh_connection,
    ):  # pylint: disable=line-too-long
        """Test that a project endpoint is created successfully."""
        mock_get_project_endpoint.return_value = False

        mock_get_project_pod.return_value = None
        with patch(
            "runpod.cli.groups.project.functions._launch_dev_pod"
        ) as mock_launch_dev_pod:
            mock_launch_dev_pod.return_value = None
            assert create_project_endpoint() is None

        mock_get_project_pod.return_value = {"id": "test_pod_id"}
        mock_load_project_config.return_value = {
            "project": {
                "name": "test_project",
                "volume_mount_path": "/runpod-volume/123456",
                "uuid": "123456",
                "env_vars": {"TEST_VAR": "value"},
                "base_image": "test_image",
                "container_disk_size_gb": 10,
                "storage_id": "test_storage_id",
            },
            "runtime": {
                "python_version": "3.8",
                "handler_path": "handler.py",
                "requirements_path": "requirements.txt",
            },
        }
        mock_create_template.return_value = {"id": "test_template_id"}
        mock_create_endpoint.return_value = {"id": "test_endpoint_id"}

        mock_ssh_instance = mock_ssh_connection.return_value
        mock_ssh_instance.__enter__.return_value = mock_ssh_instance
        mock_ssh_instance.run_commands.return_value = None

        with patch("runpod.cli.groups.project.functions.datetime") as mock_datetime:
            mock_datetime.now.return_value = "123456"
            result = create_project_endpoint()

        self.assertEqual(result, "test_endpoint_id")
        mock_create_template.assert_called_with(
            name="test_project-endpoint | 123456 | 123456",
            image_name="test_image",
            container_disk_in_gb=10,
            docker_start_cmd='bash -c ". /runpod-volume/123456/prod/venv/bin/activate && python -u /runpod-volume/123456/prod/test_project/handler.py"',  # pylint: disable=line-too-long
            env={"TEST_VAR": "value"},
            is_serverless=True,
        )
        mock_create_endpoint.assert_called_with(
            name="test_project-endpoint | 123456",
            template_id="test_template_id",
            network_volume_id="test_storage_id",
        )

        mock_update_endpoint.return_value = {"id": "test_endpoint_id"}
        mock_get_project_endpoint.return_value = {"id": "test_endpoint_id"}
        self.assertEqual(create_project_endpoint(), "test_endpoint_id")



"""RunPod | CLI | Utils | rp_info

A collection of utility functions for retrieving information about pods.
"""

import time

from runpod import get_pod


def get_pod_ssh_ip_port(pod_id, timeout=300):
    """
    Returns the IP and port for SSH access to a pod.
    """
    start_time = time.time()
    pod_ip = None
    pod_port = None

    while time.time() - start_time < timeout and (pod_ip is None or pod_port is None):
        pod = get_pod(pod_id)
        desired_status = pod.get("desiredStatus", None)
        runtime = pod.get("runtime", None)

        if desired_status == "RUNNING" and runtime and "ports" in pod["runtime"]:
            for port in pod["runtime"]["ports"]:
                if port["privatePort"] == 22:
                    pod_ip = port["ip"]
                    pod_port = int(port["publicPort"])
                    break

        time.sleep(1)

    if desired_status != "RUNNING":
        raise TimeoutError(
            f"Pod {pod_id} did not reach 'RUNNING' state within {timeout} seconds."
        )

    if runtime is None:
        raise TimeoutError(
            f"Pod {pod_id} did not report runtime data within {timeout} seconds."
        )

    return pod_ip, pod_port



"""
This module contains unit tests for the 'whatever' module, specifically the is_even function.
"""

import unittest

from .whatever import is_even


class TestIsEven(unittest.TestCase):
    """
    TestIsEven class contains test cases for the is_even function in the whatever module.
    """

    def test_is_even_true(self):
        """
        Test that is_even returns True for even numbers.
        """
        job = {"input": {"number": 2}}
        self.assertTrue(is_even(job))

    def test_is_even_false(self):
        """
        Test that is_even returns False for odd numbers.
        """
        job = {"input": {"number": 3}}
        self.assertFalse(is_even(job))

    def test_is_even_error(self):
        """
        Test that is_even returns an error message for non-integer inputs.
        """
        job = {"input": {"number": "two"}}
        expected_output = {"error": "Silly human, you need to pass an integer."}
        self.assertEqual(is_even(job), expected_output)


if __name__ == "__main__":
    unittest.main()



""" Unit testing for runpod.cli.utils.rp_info.py """

from unittest.mock import patch

import pytest

from runpod.cli.utils.rp_info import get_pod_ssh_ip_port


class TestGetPodSSHIpPort:
    """Unit testing for get_pod_ssh_ip_port"""

    def test_get_pod_ssh_ip_port_normal(self):
        """Test get_pod_ssh_ip_port normal"""
        with patch("runpod.cli.utils.rp_info.get_pod") as mock_get_pod:
            mock_get_pod.return_value = {
                "desiredStatus": "RUNNING",
                "runtime": {
                    "ports": [
                        {"privatePort": 22, "ip": "127.0.0.1", "publicPort": 2222}
                    ]
                },
            }

            ip, port = get_pod_ssh_ip_port("pod_id")
            assert ip == "127.0.0.1"
            assert port == 2222

    def test_get_pod_ssh_ip_port_timeout(self):
        """Test get_pod_ssh_ip_port timeout"""
        with patch("runpod.cli.utils.rp_info.get_pod") as mock_get_pod:
            mock_get_pod.return_value = {"desiredStatus": "RUNNING", "runtime": None}

            with pytest.raises(TimeoutError):
                get_pod_ssh_ip_port("pod_id", timeout=0.1)

            mock_get_pod.return_value = {"desiredStatus": "NOT_RUNNING"}

            with pytest.raises(TimeoutError):
                get_pod_ssh_ip_port("pod_id", timeout=0.1)

            mock_get_pod.return_value = {}

            with pytest.raises(TimeoutError):
                get_pod_ssh_ip_port("pod_id", timeout=0.1)



""" Example of streaming data from an endpoint. """

import runpod

# Set your global API key with `runpod config` or uncomment the line below:
# runpod.api_key = "YOUR_RUNPOD_API_KEY"

endpoint = runpod.Endpoint("gwp4kx5yd3nur1")

run_request = endpoint.run(
    {
        "input": {
            "mock_return": ["a", "b", "c", "d", "e", "f", "g"],
            "mock_delay": 1,
        }
    }
)

for output in run_request.stream():
    print(output)



"""
RunPod | API | Queries | GPUs
"""

QUERY_GPU_TYPES = """
query GpuTypes {
  gpuTypes {
    id
    displayName
    memoryInGb
  }
}
"""


def generate_gpu_query(gpu_id, gpu_count=1):
    """
    Generate a query for a specific GPU type
    """

    return f"""
    query GpuTypes {{
      gpuTypes(input: {{id: "{gpu_id}"}}) {{
        maxGpuCount
        id
        displayName
        manufacturer
        memoryInGb
        cudaCores
        secureCloud
        communityCloud
        securePrice
        communityPrice
        oneMonthPrice
        threeMonthPrice
        oneWeekPrice
        communitySpotPrice
        secureSpotPrice
        lowestPrice(input: {{gpuCount: {gpu_count}}}) {{
          minimumBidPrice
          uninterruptablePrice
        }}
      }}
    }}
    """



""" Tests for runpod.serverless.utils.validate """

import unittest
from unittest.mock import Mock

from runpod.serverless.utils import rp_validator


class TestValidator(unittest.TestCase):
    """Tests for validator"""

    def setUp(self):
        self.raw_input = {"a": 1.1, "x": 10, "y": 20, "z": 30}
        self.schema = {
            "a": {"type": float, "required": True},
            "x": {"type": int, "required": True},
            "y": {"type": int, "required": True, "default": 5},
            "z": {
                "type": int,
                "required": False,
                "default": 5,
                "constraints": Mock(return_value=True),
            },
            "w": {"type": int, "required": False, "default": 5},
        }

    def test_validate_success(self):
        """
        Tests validate
        """
        result = rp_validator.validate(self.raw_input, self.schema)

        self.assertNotIn("errors", result)
        expected_output = self.raw_input.copy()
        expected_output["w"] = self.schema["w"]["default"]
        self.assertEqual(result["validated_input"], expected_output)

    def test_validate_constraints_error(self):
        """
        Tests validate with constraints error
        """
        # Now add a constraint that the 'x' value must be less than 10
        self.schema["x"]["constraints"] = Mock(return_value=False)

        result = rp_validator.validate(self.raw_input, self.schema)

        self.assertIn("errors", result)
        self.assertIn("x does not meet the constraints.", result["errors"])

    def test_validate_missing_required_input(self):
        """
        Tests validate with missing required input
        """
        del self.raw_input["x"]
        result = rp_validator.validate(self.raw_input, self.schema)

        self.assertIn("errors", result)
        self.assertIn("x is a required input.", result["errors"])

    def test_validate_unexpected_input(self):
        """
        Tests validate with unexpected input
        """
        self.raw_input["unexpected"] = "unexpected"
        result = rp_validator.validate(self.raw_input, self.schema)

        self.assertIn("errors", result)
        self.assertIn(
            "Unexpected input. unexpected is not a valid input option.",
            result["errors"],
        )

    def test_validate_missing_default(self):
        """
        Tests validate with missing default
        """
        del self.schema["w"]["default"]
        result = rp_validator.validate(self.raw_input, self.schema)

        self.assertIn("errors", result)
        self.assertIn("Schema error, missing default value for w.", result["errors"])

    def test_validate_invalid_type(self):
        """
        Tests validate with invalid type
        """
        self.raw_input["x"] = "invalid"
        result = rp_validator.validate(self.raw_input, self.schema)

        self.assertIn("errors", result)
        self.assertIn(
            "x should be <class 'int'> type, not <class 'str'>.", result["errors"]
        )

    def test_validate_rules_not_dict(self):
        """
        Tests validate with rules not dict
        """
        result = rp_validator.validate(self.raw_input, {"x": "not dict"})
        self.assertIn("errors", result)

    def test_validate_simple_input(self):
        """
        Tests validate with simple input
        """
        result = rp_validator.validate(
            {"my_input": None}, {"my_input": {"type": str, "required": True}}
        )
        self.assertIn("errors", result)

    def test_validate_none_type(self):
        """
        Tests validate with None type
        """
        result = rp_validator.validate(
            {"my_input": None}, {"my_input": {"type": type(None), "required": True}}
        )
        self.assertNotIn("errors", result)


if __name__ == "__main__":
    unittest.main()



"""
Unit tests for the rp_cuda module
"""

from unittest.mock import patch

from runpod.serverless.utils import rp_cuda


def test_is_available_true():
    """
    Test that is_available returns True when nvidia-smi is available
    """
    with patch(
        "subprocess.check_output", return_value=b"NVIDIA-SMI"
    ) as mock_check_output:
        assert rp_cuda.is_available() is True
    mock_check_output.assert_called_once_with("nvidia-smi", shell=True)


def test_is_available_false():
    """
    Test that is_available returns False when nvidia-smi is not available
    """
    with patch(
        "subprocess.check_output", return_value=b"Not a GPU output"
    ) as mock_check_output:
        assert rp_cuda.is_available() is False
    mock_check_output.assert_called_once_with("nvidia-smi", shell=True)


def test_is_available_exception():
    """
    Test that is_available returns False when nvidia-smi raises an exception
    """
    with patch(
        "subprocess.check_output", side_effect=Exception("Bad Command")
    ) as mock_check:
        assert rp_cuda.is_available() is False
    mock_check.assert_called_once_with("nvidia-smi", shell=True)



"""
Tests for runpod | endpoint | modules | endpoint.py
"""

import unittest
from unittest.mock import Mock, patch

import requests

import runpod
from runpod.endpoint import runner
from runpod.endpoint.runner import Endpoint, Job, RunPodClient


class TestRunPodClient(unittest.TestCase):
    """Tests for RunPodClient"""

    def test_no_api_key(self):
        """
        Tests RunPodClient with no api_key
        """
        with self.assertRaises(RuntimeError):
            runpod.api_key = None
            RunPodClient()

    @patch.object(requests.Session, "post")
    def test_post_with_401(self, mock_post):
        """
        Tests RunPodClient.post with 401 status code
        """
        mock_response = Mock()
        mock_response.status_code = 401
        mock_post.return_value = mock_response

        with self.assertRaises(RuntimeError):
            runpod.api_key = "MOCK_API_KEY"
            client = RunPodClient()
            client.post("ENDPOINT_ID/run", {"input": {}})

    @patch.object(requests.Session, "request")
    def test_post(self, mock_post):
        """
        Tests RunPodClient.post
        """
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"id": "123"}
        mock_post.return_value = mock_response

        runpod.api_key = "MOCK_API_KEY"
        client = RunPodClient()
        response = client.post("ENDPOINT_ID/run", {"input": {}})

        self.assertEqual(response, {"id": "123"})

    @patch.object(requests.Session, "get")
    def test_get_with_401(self, mock_get):
        """
        Tests RunPodClient.get with 401 status code
        """
        mock_response = Mock()
        mock_response.status_code = 401
        mock_get.return_value = mock_response

        with self.assertRaises(RuntimeError):
            runpod.api_key = "MOCK_API_KEY"
            client = RunPodClient()
            client.get("ENDPOINT_ID/status/123")

    @patch.object(requests.Session, "request")
    def test_get(self, mock_get):
        """
        Tests RunPodClient.get
        """
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"status": "COMPLETED"}
        mock_get.return_value = mock_response

        runpod.api_key = "MOCK_API_KEY"
        client = RunPodClient()
        response = client.get("ENDPOINT_ID/status/123")

        self.assertEqual(response, {"status": "COMPLETED"})


class TestEndpoint(unittest.TestCase):
    """Tests for Endpoint"""

    ENDPOINT_ID = "ENDPOINT_ID"
    MOCK_API_KEY = "MOCK_API_KEY"
    MODEL_INPUT = {"YOUR_MODEL_INPUT_JSON": "YOUR_MODEL_INPUT_VALUE"}
    MODEL_OUTPUT = {"result": "YOUR_MODEL_OUTPUT_VALUE"}

    def setUp(self):
        """Common setup for the tests."""
        runpod.api_key = self.MOCK_API_KEY
        self.endpoint = Endpoint(self.ENDPOINT_ID)

    @patch("runpod.endpoint.runner.RunPodClient._request")
    def test_endpoint_run(self, mock_client_request):
        """Test the run method of Endpoint with a successful job initiation."""
        mock_client_request.return_value = {"id": "123", "status": "IN_PROGRESS"}

        run_request = self.endpoint.run(self.MODEL_INPUT)

        # Tests
        mock_client_request.assert_called_once_with(
            "POST",
            f"{self.ENDPOINT_ID}/run",
            {"input": {"YOUR_MODEL_INPUT_JSON": "YOUR_MODEL_INPUT_VALUE"}},
            10,
        )

        self.assertIsInstance(run_request, Job)
        self.assertEqual(run_request.job_id, "123")
        self.assertEqual(run_request.status(), "IN_PROGRESS")

        mock_client_request.assert_called_with(
            "GET", f"{self.ENDPOINT_ID}/status/123", timeout=10
        )

    @patch("runpod.endpoint.runner.RunPodClient._request")
    def test_endpoint_run_sync(self, mock_client_request):
        """Test the run_sync method of Endpoint with a successful job initiation."""
        mock_client_request.return_value = {
            "id": "123",
            "status": "COMPLETED",
            "output": self.MODEL_OUTPUT,
        }

        run_request = self.endpoint.run_sync(self.MODEL_INPUT)

        # Tests
        self.assertEqual(run_request, self.MODEL_OUTPUT)

        mock_client_request.assert_called_once_with(
            "POST",
            f"{self.ENDPOINT_ID}/runsync",
            {"input": {"YOUR_MODEL_INPUT_JSON": "YOUR_MODEL_INPUT_VALUE"}},
            86400,
        )

    @patch("runpod.endpoint.runner.RunPodClient._request")
    def test_endpoint_health(self, mock_client_request):
        """Test the health method of Endpoint"""
        self.endpoint.health()

        mock_client_request.assert_called_once_with(
            "GET", f"{self.ENDPOINT_ID}/health", timeout=3
        )

    @patch("runpod.endpoint.runner.RunPodClient._request")
    def test_endpoint_purge_queue(self, mock_client_request):
        """Test the health method of Endpoint"""
        self.endpoint.purge_queue()

        mock_client_request.assert_called_once_with(
            "POST", f"{self.ENDPOINT_ID}/purge-queue", None, 3
        )

    def test_missing_api_key(self):
        """
        Tests Endpoint.run without api_key
        """
        with self.assertRaises(RuntimeError):
            runpod.api_key = None
            self.endpoint.run(self.MODEL_INPUT)

    @patch.object(requests.Session, "post")
    def test_run_with_401(self, mock_post):
        """
        Tests Endpoint.run with 401 status code
        """
        mock_response = Mock()
        mock_response.status_code = 401
        mock_post.return_value = mock_response

        endpoint = runpod.Endpoint("ENDPOINT_ID")
        request_data = {"YOUR_MODEL_INPUT_JSON": "YOUR_MODEL_INPUT_VALUE"}

        with self.assertRaises(RuntimeError):
            endpoint.run(request_data)

    @patch.object(runpod.endpoint.runner.RunPodClient, "_request")
    def test_run(self, mock_client_request):
        """
        Tests Endpoint.run
        """
        mock_client_request.return_value = {"id": "123", "status": "IN_PROGRESS"}

        runpod.api_key = "MOCK_API_KEY"
        endpoint = runpod.Endpoint("ENDPOINT_ID")

        request_data = {"YOUR_MODEL_INPUT_JSON": "YOUR_MODEL_INPUT_VALUE"}
        run_request = endpoint.run(request_data)

        self.assertEqual(run_request.job_id, "123")
        self.assertEqual(run_request.status(), "IN_PROGRESS")

    @patch.object(runpod.endpoint.runner.RunPodClient, "_request")
    def test_run_sync(self, mock_client_request):
        """
        Tests Endpoint.run_sync
        """
        mock_client_request.return_value = {
            "id": "123",
            "status": "COMPLETED",
            "output": {"result": "YOUR_MODEL_OUTPUT_VALUE"},
        }

        runpod.api_key = "MOCK_API_KEY"
        endpoint = runpod.Endpoint("ENDPOINT_ID")

        request_data = {"YOUR_MODEL_INPUT_JSON": "YOUR_MODEL_INPUT_VALUE"}
        run_request = endpoint.run_sync(request_data)

        self.assertEqual(run_request, {"result": "YOUR_MODEL_OUTPUT_VALUE"})

    @patch.object(runpod.endpoint.runner.RunPodClient, "_request")
    def test_run_sync_with_timeout(self, mock_client_request):
        """
        Tests Endpoint.run_sync with timeout
        """
        mock_client_request.return_value = {"id": "123", "status": "IN_PROGRESS"}

        runpod.api_key = "MOCK_API_KEY"
        endpoint = runpod.Endpoint("ENDPOINT_ID")

        request_data = {"YOUR_MODEL_INPUT_JSON": "YOUR_MODEL_INPUT_VALUE"}
        with self.assertRaises(TimeoutError):
            endpoint.run_sync(request_data, timeout=1)


class TestJob(unittest.TestCase):
    """Tests for Job"""

    MODEL_OUTPUT = {"result": "YOUR_MODEL_OUTPUT_VALUE"}

    @patch("runpod.endpoint.runner.RunPodClient")
    def test_status(self, mock_client):
        """
        Tests Job.status
        """
        mock_client.get.return_value = {"status": "COMPLETED"}

        job = runner.Job("endpoint_id", "job_id", mock_client)
        status = job.status()
        self.assertEqual(status, "COMPLETED")

    @patch("runpod.endpoint.runner.RunPodClient")
    def test_output(self, mock_client):
        """
        Tests Job.output
        """
        mock_client.get.return_value = {"status": "COMPLETED", "output": "Job output"}

        job = runner.Job("endpoint_id", "job_id", mock_client)
        output = job.output()
        self.assertEqual(output, "Job output")

    @patch("runpod.endpoint.runner.RunPodClient")
    def test_output_with_sleep(self, mock_client):
        """
        Tests Job.output with sleep
        """
        mock_client.get.side_effect = [
            {"status": "IN_PROGRESS"},
            {"status": "COMPLETED", "output": "Job output"},
        ]

        job = runner.Job("endpoint_id", "job_id", mock_client)
        output = job.output(timeout=10)

        self.assertEqual(output, "Job output")

    @patch("runpod.endpoint.runner.RunPodClient")
    def test_output_timeout(self, mock_client):
        """
        Tests Job.output with timeout
        """
        mock_client.get.return_value = {"status": "IN_PROGRESS"}

        job = runner.Job("endpoint_id", "job_id", mock_client)
        with self.assertRaises(TimeoutError):
            job.output(timeout=1)

    @patch("runpod.endpoint.runner.RunPodClient")
    def test_cancel(self, mock_client):
        """Test the cancel method of Job with a successful job initiation."""
        job = runner.Job("endpoint_id", "job_id", mock_client)

        job.cancel()

        mock_client.post.assert_called_with(
            "endpoint_id/cancel/job_id", data=None, timeout=3
        )

    @patch("runpod.endpoint.runner.RunPodClient")
    def test_job_status(self, mock_client):
        """
        Tests Job.status
        """
        mock_client.get.side_effect = [
            {"status": "IN_PROGRESS"},
            {"status": "COMPLETED"},
        ]

        job = runner.Job("endpoint_id", "job_id", mock_client)
        self.assertEqual(job.status(), "IN_PROGRESS")
        self.assertEqual(job.status(), "COMPLETED")
        self.assertEqual(job.status(), "COMPLETED")

    @patch("runpod.endpoint.runner.RunPodClient")
    def test_job_stream(self, mock_client):
        """
        Tests Job.stream
        """
        mock_client.get.side_effect = [
            {
                "status": "IN_PROGRESS",
                "stream": [{"output": "Job output 1"}, {"output": "Job output 2"}],
            },
            {"status": "COMPLETED", "stream": []},
        ]

        job = runner.Job("endpoint_id", "job_id", mock_client)
        output = list(job.stream())
        self.assertEqual(output, ["Job output 1", "Job output 2"])



"""Tests for runpod.cli.utils.rp_sync module."""

import time
import unittest
from unittest.mock import ANY, MagicMock, patch

from runpod.cli import STOP_EVENT
from runpod.cli.utils.rp_sync import WatcherHandler, start_watcher, sync_directory


class TestWatcherHandler(unittest.TestCase):
    """Tests for the WatcherHandler class."""

    @patch("runpod.cli.utils.rp_sync.should_ignore")
    def test_on_any_event_with_ignored_file(self, mock_should_ignore):
        """Test that the action function is not called when the file is ignored."""
        mock_should_ignore.return_value = True
        mock_action_function = MagicMock()
        handler = WatcherHandler(mock_action_function, "some_path")

        event_mock = MagicMock()
        event_mock.is_directory = False
        event_mock.src_path = "some_path/ignored_file.txt"

        handler.on_any_event(event_mock)
        mock_action_function.assert_not_called()

    @patch("runpod.cli.utils.rp_sync.should_ignore")
    def test_on_any_event_with_not_ignored_file(self, mock_should_ignore):
        """Test that the action function is called when the file is not ignored."""
        mock_should_ignore.return_value = False
        mock_action_function = MagicMock()
        handler = WatcherHandler(mock_action_function, "some_path")

        event_mock = MagicMock()
        event_mock.is_directory = False
        event_mock.src_path = "some_path/not_ignored_file.txt"

        handler.on_any_event(event_mock)
        handler.on_any_event(event_mock)  # Call it twice to test the debouncer
        time.sleep(2)
        mock_action_function.assert_called_once()

    def test_on_any_event_with_directory(self):
        """Test that the action function is not called when the event is a directory."""
        mock_action_function = MagicMock()
        handler = WatcherHandler(mock_action_function, "some_path")

        event_mock = MagicMock()
        event_mock.is_directory = True

        handler.on_any_event(event_mock)
        time.sleep(2)
        mock_action_function.assert_not_called()


class TestSyncDirectory(unittest.TestCase):
    """Tests for the sync_directory function."""

    @patch("runpod.cli.utils.rp_sync.threading.Thread")
    @patch("runpod.cli.utils.rp_sync.start_watcher")
    def test_sync_directory(self, mock_start_watcher, mock_thread_class):
        """Test that the sync_directory function calls the start_watcher function."""
        mock_ssh_client = MagicMock()

        local_path = "local_path"
        remote_path = "remote_path"

        sync_directory(mock_ssh_client, local_path, remote_path)

        target_function = mock_thread_class.call_args[1]["target"]
        target_function()

        mock_start_watcher.assert_called_once()

    @patch("runpod.cli.utils.rp_sync.threading.Thread")
    @patch("runpod.cli.utils.rp_sync.start_watcher")
    def test_sync_directory_sync_function(self, mock_start_watcher, mock_thread_class):
        """Test that the sync_directory function calls the start_watcher function."""
        mock_ssh_client = MagicMock()

        local_path = "local_path"
        remote_path = "remote_path"

        sync_function = sync_directory(mock_ssh_client, local_path, remote_path)
        sync_function()

        mock_ssh_client.rsync.assert_called_once_with(
            local_path, remote_path, quiet=True
        )

        mock_thread_class.assert_called_once()
        mock_thread_class.assert_called_with(
            target=mock_start_watcher, daemon=True, args=(ANY, local_path)
        )

        assert mock_start_watcher.called is False


class TestStartWatcher(unittest.TestCase):
    """Tests for the start_watcher function."""

    @patch("runpod.cli.utils.rp_sync.Observer")
    @patch("runpod.cli.utils.rp_sync.WatcherHandler")
    def test_start_watcher(self, mock_watch_handler, mock_observer_class):
        """Test that the start_watcher function starts the watcher correctly."""
        fake_action = MagicMock()
        local_path = "/path/to/watch"

        mock_observer_instance = mock_observer_class.return_value

        STOP_EVENT.clear()
        with patch("runpod.cli.utils.rp_sync.time.sleep") as mock_sleep:

            def side_effect(*args, **kwargs):
                del args, kwargs
                STOP_EVENT.set()

            mock_sleep.side_effect = side_effect
            start_watcher(fake_action, local_path)

        mock_watch_handler.assert_called_once_with(fake_action, local_path)

        mock_observer_instance.schedule.assert_called_once_with(
            mock_watch_handler.return_value, local_path, recursive=True
        )

        mock_observer_instance.start.assert_called_once()
        mock_observer_instance.stop.assert_called_once()
        mock_observer_instance.join.assert_called_once()



"""Helper functions for project group commands."""

import os
import re
import shutil

import click
import tomlkit

from runpod import create_pod
from runpod import error as rp_error
from runpod import get_endpoints, get_pods


def validate_project_name(name):
    """
    Validate the project name.
    """
    match = re.search(r"[<>:\"/\\|?*\s]", name)
    if match:
        raise click.BadParameter(
            f"Project name contains an invalid character: '{match.group()}'."
        )
    return name


def get_project_pod(project_id: str):
    """Check if a project pod exists.
    Return the pod_id if it exists, else return None.
    """
    for pod in get_pods():
        if project_id in pod["name"]:
            return pod["id"]

    return None


def get_project_endpoint(project_id: str):
    """Check if a project endpoint exists.
    Return the endpoint if it exists, else return None.
    """
    for endpoint in get_endpoints():
        if project_id in endpoint["name"]:
            return endpoint

    return None


def copy_template_files(template_dir, destination):
    """Copy the template files to the destination directory."""
    for item in os.listdir(template_dir):
        source_item = os.path.join(template_dir, item)
        destination_item = os.path.join(destination, item)
        if os.path.isdir(source_item):
            shutil.copytree(source_item, destination_item)
        else:
            shutil.copy2(source_item, destination_item)


def attempt_pod_launch(config, environment_variables):
    """Attempt to launch a pod with the given configuration."""
    for gpu_type in config["project"].get("gpu_types", []):
        print(f"Trying to get a pod with {gpu_type}... ", end="")
        try:
            created_pod = create_pod(
                f'{config["project"]["name"]}-dev ({config["project"]["uuid"]})',
                config["project"]["base_image"],
                gpu_type,
                gpu_count=int(config["project"]["gpu_count"]),
                support_public_ip=True,
                ports=f'{config["project"]["ports"]}',
                network_volume_id=f'{config["project"]["storage_id"]}',
                volume_mount_path=f'{config["project"]["volume_mount_path"]}',
                container_disk_in_gb=int(config["project"]["container_disk_size_gb"]),
                env=environment_variables,
            )
            print("Success!")
            return created_pod
        except rp_error.QueryError:
            print("Unavailable.")
    return None


def load_project_config():
    """Load the project config file."""
    project_config_file = os.path.join(os.getcwd(), "runpod.toml")
    if not os.path.exists(project_config_file):
        raise FileNotFoundError("runpod.toml not found in the current directory.")
    with open(project_config_file, "r", encoding="UTF-8") as config_file:
        return tomlkit.load(config_file)



"""
RunPod Tips
"""

import sys

import runpod.serverless.modules.rp_logger as RunPodLogger

log = RunPodLogger.RunPodLogger()


def check_return_size(return_body):
    """
    Checks the size of the return body.
    If the size is above 20MB, it will recommend using storage upload.
    """
    size_bytes = sys.getsizeof(return_body)
    size_mb = round(size_bytes / 1_000_000, 2)

    if size_mb > 20:
        log.tip(
            f"Your return body is {size_mb} MB which exceeds the 20 MB limit. "
            "Consider using S3 upload and returning the object's URL instead."
        )



"""
RunPod Progress Module
"""

import asyncio
import threading
from typing import Any, Dict

from runpod.http_client import AsyncClientSession
from runpod.serverless.modules.rp_logger import RunPodLogger

from .rp_http import send_result

log = RunPodLogger()


async def _async_progress_update(session, job, progress):
    """
    The actual asynchronous function that sends the update.
    """
    job_data = {"status": "IN_PROGRESS", "output": progress}

    await send_result(session, job_data, job)


def _thread_target(job: Dict[str, Any], progress: Any):
    """
    A wrapper around _async_progress_update to handle the event loop.
    """
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:

        async def main():
            session = AsyncClientSession()
            async with session:
                await _async_progress_update(session, job, progress)

        loop.run_until_complete(main())

        log.debug(f'{job["id"]} | Progress Update Sent: {progress}')
    finally:
        loop.close()


def progress_update(job: Dict[str, Any], progress: Any) -> None:
    """
    Updates the progress of a currently running job in a separate thread.
    """
    log.debug(f'{job["id"]} | Sending Progress Update: {progress}')
    thread = threading.Thread(target=_thread_target, args=(job, progress), daemon=True)
    thread.start()



""" Tests for runpod.serverless.modules.rp_logger """

import os
import unittest
from unittest.mock import patch

from runpod.serverless.modules import rp_logger


class TestLogger(unittest.TestCase):
    """Tests for rp_logger"""

    def setUp(self) -> None:
        """
        Set up the logger for each test
        """
        self.logger = rp_logger.RunPodLogger()

    def test_default_log_level(self):
        """
        Tests that the default log level is DEBUG
        """
        defult_logger = rp_logger.RunPodLogger()

        self.assertEqual(defult_logger.level, "DEBUG")

    def test_singleton(self):
        """
        Tests that the logger is a singleton
        """
        logger1 = rp_logger.RunPodLogger()
        logger2 = rp_logger.RunPodLogger()

        self.assertIs(logger1, logger2)

    def test_set_log_level(self):
        """
        Tests that the log level can be set
        """
        logger = rp_logger.RunPodLogger()

        logger.set_level("TRACE")
        self.assertEqual(logger.level, "TRACE")

        logger.set_level("INFO")
        self.assertEqual(logger.level, "INFO")

        logger.set_level("WARN")
        self.assertEqual(logger.level, "WARN")

        logger.set_level(3)
        self.assertEqual(logger.level, "INFO")

    def test_call_log(self):
        """
        Tests that the logger can be called and logs the message to stdout if the log level is set.
        """
        log = rp_logger.RunPodLogger()

        with patch("runpod.serverless.modules.rp_logger.RunPodLogger.log") as mock_log:

            log.warn("Test log message")

            mock_log.assert_called_once_with("Test log message", "WARN", None)

        log.set_level(0)
        with patch(
            "runpod.serverless.modules.rp_logger.RunPodLogger.log"
        ) as mock_log, patch("builtins.print") as mock_print:

            log.debug("Test log message")

            mock_log.assert_called_once_with("Test log message", "DEBUG", None)
            mock_print.assert_not_called()

        # Reset log level
        log.set_level("DEBUG")

    def test_invalid_debug_level(self):
        """
        Tests that an invalid debug level raises an exception
        """
        logger = rp_logger.RunPodLogger()

        with self.assertRaises(ValueError):
            logger.set_level("INVALID")

        with self.assertRaises(ValueError):
            logger.set_level([])

    def test_debug_level_int(self):
        """
        Tests that the debug level can be set using an int
        """
        int_logger = rp_logger.RunPodLogger()

        with self.assertRaises(ValueError):
            int_logger.set_level(69)

    def test_log_secret(self):
        """
        Tests that the secret method censors secrets.
        Captures stdout and checks that the secret is censored.
        """
        with patch("runpod.serverless.modules.rp_logger.RunPodLogger.log") as mock_log:
            self.logger.secret("test_secret", "test_secret_value")
            mock_log.assert_called_once_with(
                "test_secret: t***************e", "INFO", None
            )

    def test_log_tip(self):
        """
        Tests that the tip method logs a tip.
        """
        with patch("runpod.serverless.modules.rp_logger.RunPodLogger.log") as mock_log:
            self.logger.tip("test_tip")
            mock_log.assert_called_once_with("test_tip", "TIP")

    def test_log_trace(self):
        """
        Tests that the trace method logs a trace.
        """
        with patch("runpod.serverless.modules.rp_logger.RunPodLogger.log") as mock_log:
            self.logger.trace("This is a trace message")
            mock_log.assert_called_once_with("This is a trace message", "TRACE", None)

            self.logger.trace("This is another trace message", "test-request-id")
            mock_log.assert_called_with(
                "This is another trace message", "TRACE", "test-request-id"
            )

    def test_log_job_id(self):
        """Tests that the log method logs a job id"""
        logger = rp_logger.RunPodLogger()
        job_id = "test_job_id"

        # Patch print to capture stdout
        with patch("builtins.print") as mock_print:
            logger.log("test_message", "INFO", job_id)

            mock_print.assert_called_once_with(
                "INFO   | test_job_id | test_message", flush=True
            )

            # Test with endpoint id set
            os.environ["RUNPOD_ENDPOINT_ID"] = "test_endpoint_id"
            logger.log("test_message", "INFO", job_id)
            os.environ.pop("RUNPOD_ENDPOINT_ID")

            mock_print.assert_called_with(
                '{"requestId": "test_job_id", "message": "test_message", "level": "INFO"}',
                flush=True,
            )

    def test_log_truncate(self):
        """Tests that the log method truncates"""
        logger = rp_logger.RunPodLogger()
        job_id = "test_job_id"
        long_message = "a" * (rp_logger.MAX_MESSAGE_LENGTH + 100)
        expected_start = "a" * (rp_logger.MAX_MESSAGE_LENGTH // 2)
        expected_end = "a" * (rp_logger.MAX_MESSAGE_LENGTH // 2)
        truncated_amount = len(long_message) - rp_logger.MAX_MESSAGE_LENGTH
        truncation_note = f"\n...TRUNCATED {truncated_amount} CHARACTERS...\n"
        truncated_message = expected_start + truncation_note + expected_end

        with patch("builtins.print") as mock_print:
            logger.log(long_message, "INFO", job_id)

            expected_log_output = f"INFO   | {job_id} | {truncated_message}"

            mock_print.assert_called_once_with(expected_log_output, flush=True)



"""
Example of calling an endpoint using asyncio.
"""

import asyncio

import runpod
from runpod import AsyncioEndpoint, AsyncioJob, http_client

asyncio.set_event_loop_policy(
    asyncio.WindowsSelectorEventLoopPolicy()
)  # For Windows Users

runpod.api_key = "YOUR_API_KEY"


async def main():
    """
    Function to run the example.
    """
    async with http_client.AsyncClientSession() as session:
        # Invoke API
        payload = {}
        endpoint = AsyncioEndpoint("ENDPOINT_ID", session)
        job: AsyncioJob = await endpoint.run(payload)

        # Get current job status
        status = await job.status()

        # Print status
        print(status)

        # Wait until job is completed or failed
        output = await job.output()

        # Print output
        print(output)


asyncio.run(main())



""" Concurrency Modifier Example """

import runpod


# ---------------------------------- Handler --------------------------------- #
async def async_generator_handler(job):
    """
    Async generator type handler.
    """
    return job


# --------------------------- Concurrency Modifier --------------------------- #
def concurrency_modifier(current_concurrency=1):
    """
    Concurrency modifier.
    """
    desired_concurrency = current_concurrency

    # Do some logic to determine the desired concurrency.

    return desired_concurrency


runpod.serverless.start(
    {"handler": async_generator_handler, "concurrency_modifier": concurrency_modifier}
)



""" Tests for the SSH functions """

import base64
import unittest
from unittest.mock import mock_open, patch

from runpod.cli.groups.ssh.functions import (
    add_ssh_key,
    generate_ssh_key_pair,
    get_ssh_key_fingerprint,
    get_user_pub_keys,
)


class TestSSHFunctions(unittest.TestCase):
    """Tests for the SSH functions"""

    def test_get_ssh_key_fingerprint(self):
        """Test the get_ssh_key_fingerprint function"""
        key = "ssh-rsa AAAAB3Nza...base64data...Q== user@host"
        fingerprint = get_ssh_key_fingerprint(key)
        expected_start = "SHA256:"
        self.assertTrue(fingerprint.startswith(expected_start))

    def test_get_ssh_key_fingerprint_invalid(self):
        """Test the get_ssh_key_fingerprint function with an invalid key"""
        with self.assertRaises(ValueError):
            get_ssh_key_fingerprint("invalidkey")

    @patch("runpod.cli.groups.ssh.functions.get_user")
    def test_get_user_pub_keys(self, mock_get_user):
        """Test the get_user_pub_keys function"""

        # Create dummy base64 data for our mock SSH keys
        dummy_data1 = base64.b64encode("test data 1".encode("utf-8")).decode("utf-8")
        dummy_data2 = base64.b64encode("test data 2".encode("utf-8")).decode("utf-8")

        mock_get_user.return_value = {
            "pubKey": f"ssh-rsa {dummy_data1} key1\nssh-rsa {dummy_data2} key2\n1"
        }

        keys = get_user_pub_keys()

        self.assertEqual(len(keys), 2)
        self.assertEqual(keys[0]["fingerprint"].startswith("SHA256:"), True)

    @patch("runpod.cli.groups.ssh.functions.get_user")
    def test_add_ssh_key_already_exists(self, mock_get_user):
        """Test the add_ssh_key function when the key already exists"""
        mock_get_user.return_value = {"pubKey": "ssh-rsa ABCDE12345 key1"}
        key = "ssh-rsa AAAAB3Nza...base64data...Q== user@host"
        with patch(
            "runpod.cli.groups.ssh.functions.update_user_settings"
        ) as mock_update_settings:
            mock_update_settings.return_value = None
            self.assertIsNone(add_ssh_key(key))
            assert mock_update_settings.called

    @patch("runpod.cli.groups.ssh.functions.os.path.join")
    @patch("runpod.cli.groups.ssh.functions.paramiko.RSAKey.generate")
    def test_generate_ssh_key_pair(self, mock_generate, mock_path_join):
        """Test the generate_ssh_key_pair function"""
        mock_generate.return_value.get_name.return_value = "ssh-rsa"
        mock_generate.return_value.get_base64.return_value = "ABCDE12345"
        mock_path_join.return_value = "/path/to/private_key"

        with patch("os.mkdir") as mock_mkdir, patch(
            "builtins.open", mock_open()
        ) as mock_file, patch(
            "runpod.cli.groups.ssh.functions.os.chmod"
        ) as mock_chmod, patch(
            "runpod.cli.groups.ssh.functions.add_ssh_key"
        ) as mock_add_key:
            mock_mkdir.return_value = None
            mock_file.return_value.write.return_value = None
            private_key, public_key = generate_ssh_key_pair("test_key")
            self.assertEqual(public_key, "ssh-rsa ABCDE12345 test_key")
            assert private_key is not None
            assert mock_file.called
            assert mock_add_key.called
            assert mock_chmod.called

    @patch("runpod.cli.groups.ssh.functions.get_user")
    @patch("runpod.cli.groups.ssh.functions.update_user_settings")
    def test_add_ssh_key_new(self, mock_update_settings, mock_get_user):
        """Test the add_ssh_key function when the key is new"""
        mock_get_user.return_value = {"pubKey": ""}
        key = "ssh-rsa ABCDE12345 somecomment"
        add_ssh_key(key)
        mock_update_settings.assert_called_once_with(key)

        mock_get_user.return_value = {"pubKey": "ssh-rsa ABCDE12345 key1"}
        assert add_ssh_key("ssh-rsa ABCDE12345 key1") is None



""" A template for a handler file. """

import runpod


def handler(job):
    """
    This is the handler function for the job.
    """
    job_input = job["input"]
    name = job_input.get("name", "World")
    return f"Hello, {name}!"


runpod.serverless.start({"handler": handler})



""" Simple Handler

To setup a local API server, run the following command:
python simple_handler.py --rp_serve_api
"""

import runpod


def handler(job):
    """Simple handler"""
    job_input = job["input"]

    return f"Hello {job_input['name']}!"


runpod.serverless.start({"handler": handler})



"""
Commands for the config command group
"""

import sys

import click

from .functions import check_credentials, set_credentials


@click.command("config", help="Configures the RunPod CLI with the user's API key.")
@click.argument("api-key", required=False, default=None)
@click.option(
    "--profile", default="default", help="The profile to set the credentials for."
)
@click.option("--check", is_flag=True, help="Check if credentials are already set.")
def config_wizard(api_key, profile, check):
    """Starts the configuration wizard to set up the RunPod CLI.
    If credentials are already set, prompts the user to overwrite them.
    """
    valid, error = check_credentials(profile)

    if check and valid:
        click.echo("Credentials already set for profile: " + profile)
        sys.exit(0)
    elif check and not valid:
        click.echo(error)
        sys.exit(1)

    if valid:
        click.confirm(
            f"Credentials already set for profile: {profile}. Overwrite?", abort=True
        )

    if api_key is None:
        click.echo("Please enter your RunPod API Key.")
        click.echo("You can find it at https://www.runpod.io/console/user/settings")
        api_key = click.prompt(
            "    > RunPod API Key", hide_input=False, confirmation_prompt=False
        )

    set_credentials(api_key, profile, overwrite=True)

    click.echo(f"Credentials set for profile: {profile} in ~/.runpod/config.toml")



"""
runpod | serverless | cleanup.py

Called to clean up the worker pod after a job is completed.
"""

import os
import shutil
from typing import List


def clean(folder_list: List[str] = None):
    """
    Removes the downloads folder.
    """
    shutil.rmtree("input_objects", ignore_errors=True)
    shutil.rmtree("output_objects", ignore_errors=True)

    shutil.rmtree("job_files", ignore_errors=True)

    if os.path.exists("output.zip"):
        os.remove("output.zip")

    if folder_list is not None:
        for folder in folder_list:
            shutil.rmtree(folder, ignore_errors=True)



"""Module for running endpoints asynchronously."""

# pylint: disable=too-few-public-methods,R0801

import asyncio
from typing import Any, Dict

from runpod.endpoint.helpers import FINAL_STATES, is_completed
from runpod.http_client import ClientSession


class Job:
    """Class representing a job for an asynchronous endpoint"""

    def __init__(self, endpoint_id: str, job_id: str, session: ClientSession):
        from runpod import (  # pylint: disable=import-outside-toplevel,cyclic-import
            api_key,
            endpoint_url_base,
        )

        self.endpoint_id = endpoint_id
        self.job_id = job_id
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
            "X-Request-ID": job_id,
        }
        self.session = session
        self.endpoint_url_base = endpoint_url_base

        self.job_status = None
        self.job_output = None

    async def _fetch_job(self, source: str = "status") -> Dict[str, Any]:
        """Returns the raw json of the status, reaises an exception if invalid.

        Args:
            source: The URL source path of the job status.
        """
        status_url = (
            f"{self.endpoint_url_base}/{self.endpoint_id}/{source}/{self.job_id}"
        )
        job_state = await self.session.get(status_url, headers=self.headers)
        job_state = await job_state.json()

        if is_completed(job_state["status"]):
            self.job_status = job_state["status"]
            self.job_output = job_state.get("output", None)

        return job_state

    async def status(self) -> str:
        """Gets jobs' status

        Returns:
            COMPLETED, FAILED or IN_PROGRESS
        """
        if self.job_status is not None:
            return self.job_status

        job_state = await self._fetch_job()
        return job_state["status"]

    async def _wait_for_completion(self):
        while not is_completed(await self.status()):
            await asyncio.sleep(1)

    async def output(self, timeout: int = 0) -> Any:
        """Waits for serverless API job to complete or fail

        Returns:
            Output of job
        Raises:
            KeyError if job Failed
        """
        if self.job_output is not None:
            return self.job_output

        try:
            await asyncio.wait_for(self._wait_for_completion(), timeout)
        except asyncio.TimeoutError as exc:
            raise TimeoutError("Job timed out.") from exc

        job_data = await self._fetch_job()
        return job_data.get("output", None)

    async def stream(self) -> Any:
        """Returns a generator that yields the output of the job request."""
        while True:
            await asyncio.sleep(1)
            stream_partial = await self._fetch_job(source="stream")
            if (
                stream_partial["status"] not in FINAL_STATES
                or len(stream_partial.get("stream", [])) > 0
            ):
                for chunk in stream_partial.get("stream", []):
                    yield chunk["output"]
            elif stream_partial["status"] in FINAL_STATES:
                break

    async def cancel(self) -> dict:
        """Cancels current job

        Returns:
            Output of cancel operation
        """
        cancel_url = f"{self.endpoint_url_base}/{self.endpoint_id}/cancel/{self.job_id}"
        async with self.session.post(cancel_url, headers=self.headers) as resp:
            return await resp.json()


class Endpoint:
    """Class for running endpoint"""

    def __init__(self, endpoint_id: str, session: ClientSession):
        from runpod import (
            api_key,  # pylint: disable=import-outside-toplevel
            endpoint_url_base,
        )

        self.endpoint_id = endpoint_id
        self.endpoint_url = f"{endpoint_url_base}/{self.endpoint_id}/run"
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }
        self.session = session

    async def run(self, endpoint_input: dict) -> Job:
        """Runs endpoint with specified input

        Args:
            endpoint_input: any dictionary with input

        Returns:
            Newly created job
        """
        async with self.session.post(
            self.endpoint_url, headers=self.headers, json={"input": endpoint_input}
        ) as resp:
            json_resp = await resp.json()

        return Job(self.endpoint_id, json_resp["id"], self.session)

    async def health(self) -> dict:
        """Checks health of endpoint

        Returns:
            Health of endpoint
        """
        async with self.session.get(
            f"{self.endpoint_id}/health", headers=self.headers
        ) as resp:
            return await resp.json()

    async def purge_queue(self) -> dict:
        """Purges queue of endpoint

        Returns:
            Purge status
        """
        async with self.session.post(
            f"{self.endpoint_id}/purge", headers=self.headers
        ) as resp:
            return await resp.json()



"""
runpod | serverless | rp_scale.py
Provides the functionality for scaling the runpod serverless worker.
"""

import asyncio
import signal
import sys
import traceback
from typing import Any, Dict

from ...http_client import AsyncClientSession, ClientSession, TooManyRequests
from .rp_job import get_job, handle_job
from .rp_logger import RunPodLogger
from .worker_state import JobsProgress, IS_LOCAL_TEST

log = RunPodLogger()
job_progress = JobsProgress()


def _handle_uncaught_exception(exc_type, exc_value, exc_traceback):
    exc = traceback.format_exception(exc_type, exc_value, exc_traceback)
    log.error(f"Uncaught exception | {exc}")


def _default_concurrency_modifier(current_concurrency: int) -> int:
    """
    Default concurrency modifier.

    This function returns the current concurrency without any modification.

    Args:
        current_concurrency (int): The current concurrency.

    Returns:
        int: The current concurrency.
    """
    return current_concurrency


class JobScaler:
    """
    Job Scaler. This class is responsible for scaling the number of concurrent requests.
    """

    def __init__(self, config: Dict[str, Any]):
        self._shutdown_event = asyncio.Event()
        self.current_concurrency = 1
        self.config = config

        self.jobs_queue = asyncio.Queue(maxsize=self.current_concurrency)

        self.concurrency_modifier = _default_concurrency_modifier
        self.jobs_fetcher = get_job
        self.jobs_fetcher_timeout = 90
        self.jobs_handler = handle_job

        if concurrency_modifier := config.get("concurrency_modifier"):
            self.concurrency_modifier = concurrency_modifier

        if not IS_LOCAL_TEST:
            # below cannot be changed unless local
            return

        if jobs_fetcher := self.config.get("jobs_fetcher"):
            self.jobs_fetcher = jobs_fetcher

        if jobs_fetcher_timeout := self.config.get("jobs_fetcher_timeout"):
            self.jobs_fetcher_timeout = jobs_fetcher_timeout

        if jobs_handler := self.config.get("jobs_handler"):
            self.jobs_handler = jobs_handler

    async def set_scale(self):
        self.current_concurrency = self.concurrency_modifier(self.current_concurrency)

        if self.jobs_queue and (self.current_concurrency == self.jobs_queue.maxsize):
            # no need to resize
            return

        while self.current_occupancy() > 0:
            # not safe to scale when jobs are in flight
            await asyncio.sleep(1)
            continue

        self.jobs_queue = asyncio.Queue(maxsize=self.current_concurrency)
        log.debug(
            f"JobScaler.set_scale | New concurrency set to: {self.current_concurrency}"
        )

    def start(self):
        """
        This is required for the worker to be able to shut down gracefully
        when the user sends a SIGTERM or SIGINT signal. This is typically
        the case when the worker is running in a container.
        """
        sys.excepthook = _handle_uncaught_exception

        try:
            # Register signal handlers for graceful shutdown
            signal.signal(signal.SIGTERM, self.handle_shutdown)
            signal.signal(signal.SIGINT, self.handle_shutdown)
        except ValueError:
            log.warning("Signal handling is only supported in the main thread.")

        # Start the main loop
        # Run forever until the worker is signalled to shut down.
        asyncio.run(self.run())

    def handle_shutdown(self, signum, frame):
        """
        Called when the worker is signalled to shut down.

        This function is called when the worker receives a signal to shut down, such as
        SIGTERM or SIGINT. It sets the shutdown event, which will cause the worker to
        exit its main loop and shut down gracefully.

        Args:
            signum: The signal number that was received.
            frame: The current stack frame.
        """
        log.debug(f"Received shutdown signal: {signum}.")
        self.kill_worker()

    async def run(self):
        # Create an async session that will be closed when the worker is killed.
        async with AsyncClientSession() as session:
            # Create tasks for getting and running jobs.
            jobtake_task = asyncio.create_task(self.get_jobs(session))
            jobrun_task = asyncio.create_task(self.run_jobs(session))

            tasks = [jobtake_task, jobrun_task]

            # Concurrently run both tasks and wait for both to finish.
            await asyncio.gather(*tasks)

    def is_alive(self):
        """
        Return whether the worker is alive or not.
        """
        return not self._shutdown_event.is_set()

    def kill_worker(self):
        """
        Whether to kill the worker.
        """
        log.debug("Kill worker.")
        self._shutdown_event.set()

    def current_occupancy(self) -> int:
        current_queue_count = self.jobs_queue.qsize()
        current_progress_count = job_progress.get_job_count()

        log.debug(
            f"JobScaler.status | concurrency: {self.current_concurrency}; queue: {current_queue_count}; progress: {current_progress_count}"
        )
        return current_progress_count + current_queue_count

    async def get_jobs(self, session: ClientSession):
        """
        Retrieve multiple jobs from the server in batches using blocking requests.

        Runs the block in an infinite loop while the worker is alive.

        Adds jobs to the JobsQueue
        """
        while self.is_alive():
            await self.set_scale()

            jobs_needed = self.current_concurrency - self.current_occupancy()
            if jobs_needed <= 0:
                log.debug("JobScaler.get_jobs | Queue is full. Retrying soon.")
                await asyncio.sleep(1)  # don't go rapidly
                continue

            try:
                log.debug("JobScaler.get_jobs | Starting job acquisition.")

                # Keep the connection to the blocking call with timeout
                acquired_jobs = await asyncio.wait_for(
                    self.jobs_fetcher(session, jobs_needed),
                    timeout=self.jobs_fetcher_timeout,
                )

                if not acquired_jobs:
                    log.debug("JobScaler.get_jobs | No jobs acquired.")
                    continue

                for job in acquired_jobs:
                    await self.jobs_queue.put(job)
                    job_progress.add(job)
                    log.debug("Job Queued", job["id"])

                log.info(f"Jobs in queue: {self.jobs_queue.qsize()}")

            except TooManyRequests:
                log.debug(
                    f"JobScaler.get_jobs | Too many requests. Debounce for 5 seconds."
                )
                await asyncio.sleep(5)  # debounce for 5 seconds
            except asyncio.CancelledError:
                log.debug("JobScaler.get_jobs | Request was cancelled.")
                raise  # CancelledError is a BaseException
            except TimeoutError:
                log.debug("JobScaler.get_jobs | Job acquisition timed out. Retrying.")
            except TypeError as error:
                log.debug(f"JobScaler.get_jobs | Unexpected error: {error}.")
            except Exception as error:
                log.error(
                    f"Failed to get job. | Error Type: {type(error).__name__} | Error Message: {str(error)}"
                )
            finally:
                # Yield control back to the event loop
                await asyncio.sleep(0)

    async def run_jobs(self, session: ClientSession):
        """
        Retrieve jobs from the jobs queue and process them concurrently.

        Runs the block in an infinite loop while the worker is alive or jobs queue is not empty.
        """
        tasks = []  # Store the tasks for concurrent job processing

        while self.is_alive() or not self.jobs_queue.empty():
            # Fetch as many jobs as the concurrency allows
            while len(tasks) < self.current_concurrency and not self.jobs_queue.empty():
                job = await self.jobs_queue.get()

                # Create a new task for each job and add it to the task list
                task = asyncio.create_task(self.handle_job(session, job))
                tasks.append(task)

            # Wait for any job to finish
            if tasks:
                log.info(f"Jobs in progress: {len(tasks)}")

                done, pending = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED
                )

                # Remove completed tasks from the list
                tasks = [t for t in tasks if t not in done]

            # Yield control back to the event loop
            await asyncio.sleep(0)

        # Ensure all remaining tasks finish before stopping
        await asyncio.gather(*tasks)

    async def handle_job(self, session: ClientSession, job: dict):
        """
        Process an individual job. This function is run concurrently for multiple jobs.
        """
        try:
            log.debug("Handling Job", job["id"])

            await self.jobs_handler(session, self.config, job)

            if self.config.get("refresh_worker", False):
                self.kill_worker()

        except Exception as err:
            log.error(f"Error handling job: {err}", job["id"])
            raise err

        finally:
            # Inform Queue of a task completion
            self.jobs_queue.task_done()

            # Job is no longer in progress
            job_progress.remove(job)

            log.debug("Finished Job", job["id"])



"""
Example of calling an endpoint using the RunPod Python Language Library.
"""

import runpod

# Set your global API key with `runpod config` or uncomment the line below:
# runpod.api_key = "YOUR_RUNPOD_API_KEY"

endpoint = runpod.Endpoint("sdxl")  # Where "sdxl" is the endpoint ID

run_request = endpoint.run(
    {"input": {"prompt": "a photo of a horse the size of a Boeing 787"}}
)

# Check the status of the run request
print(run_request.status())

# Get the output of the endpoint run request.
print(run_request.output())

# Get the output of the endpoint run request.
# If timeout is greater than 0, this will block until the endpoint run is complete.
print(run_request.output(timeout=60))



import asyncio
import math
from faker import Faker
from typing import Any, Dict, Optional, List

from runpod.serverless.modules.rp_scale import JobScaler, RunPodLogger, JobsProgress

fake = Faker()
log = RunPodLogger()
job_progress = JobsProgress()


# Change this number to your desired concurrency
start = 1


# sample concurrency modifier that loops
def collatz_conjecture(current_concurrency):
    if current_concurrency == 1:
        return start

    if current_concurrency % 2 == 0:
        return math.floor(current_concurrency / 2)
    else:
        return current_concurrency * 3 + 1


def fake_job():
    # Change this number to your desired delay
    delay = fake.random_digit_above_two()
    return {
        "id": fake.uuid4(),
        "input": fake.sentence(),
        "mock_delay": delay,
    }


async def fake_get_job(session, num_jobs: int = 1) -> Optional[List[Dict[str, Any]]]:
    # Change this number to your desired delay
    delay = fake.random_digit_above_two() - 1

    log.info(f"... artificial delay ({delay}s)")
    await asyncio.sleep(delay)  # Simulates a blocking process

    jobs = [fake_job() for _ in range(num_jobs)]
    log.info(f"... Generated # jobs: {len(jobs)}")
    return jobs


async def fake_handle_job(session, config, job) -> dict:
    await asyncio.sleep(job["mock_delay"])  # Simulates a blocking process
    log.info(f"... Job handled ({job['mock_delay']}s)", job["id"])


job_scaler = JobScaler(
    {
        # "concurrency_modifier": collatz_conjecture,
        # "jobs_fetcher_timeout": 5,
        "jobs_fetcher": fake_get_job,
        "jobs_handler": fake_handle_job,
    }
)
job_scaler.start()
